class AllowedMethod:
    GET = 'GET'
    POST = 'POST'
    PATCH = 'PATCH'
    PUT = 'PUT'
    DELETE = 'DELETE'


class FieldType:
    STRING = 'String'
    CHARACTER = 'Character'
    TEXT = 'Text'
    INTEGER = 'Integer'
    FLOAT = 'Float'
    POSITIVE_INTEGER = 'PositiveInteger'
    NEGATIVE_INTEGER = 'NegativeInteger'
    POSITIVE_FLOAT = 'PositiveFloat'
    NEGATIVE_FLOAT = 'NegativeFloat'
    DATE = 'Date'
    DATETIME = 'DateTime'
    TIME = 'Time'
    FILE = 'File'
    IMAGE = 'Image'
    ID = 'ID'
    UUID = 'UUID'
    LIST = 'List'
    DICTIONARY = 'Dictionary'
    BOOLEAN = 'Boolean'
    UNKNOWN = 'Unknown'


class PresentmentType:
    OPTIONAL = 'Optional'
    REQUIRED = 'Required'


class Helper:
    def __init__(
        self,
        allowed_methods: list = None,
        request_fields: dict = None,
        query_params: dict = None,
        expected_response_data_fields: dict = None,
        headers_keys: dict = None,
        cookies_keys: dict = None,
    ):
        self.allowed_methods = allowed_methods or []
        self.request_fields = request_fields or {}
        self.query_params = query_params or {}
        self.expected_response_data_fields = expected_response_data_fields or {}
        self.headers_keys = headers_keys or []
        self.cookies_keys = cookies_keys or []

    _allowed_methods = None

    @property
    def allowed_methods(self):
        return {
            'key': 'allowed_methods',
            'value': self._allowed_methods,
        }

    @allowed_methods.setter
    def allowed_methods(self, value):
        self._allowed_methods = value

    _request_fields = None

    @property
    def request_fields(self):
        value = {}
        for k, v in self._request_fields.items():
            value[k] = v.todict()
        return {
            'key': 'request_fields',
            'value': value,
        }

    @request_fields.setter
    def request_fields(self, value):
        self._request_fields = value

    _query_params = None

    @property
    def query_params(self):
        return {
            'key': 'query_params',
            'value': self._query_params.todict(),
        }

    @query_params.setter
    def query_params(self, value):
        self._query_params = value

    _expected_response_data_fields = None

    @property
    def expected_response_data_fields(self):
        value = {}
        for k, v in self._expected_response_data_fields.items():
            value[k] = v.todict()
        return {
            'key': 'expected_response_data_fields',
            'value': value,
        }

    @expected_response_data_fields.setter
    def expected_response_data_fields(self, value):
        self._expected_response_data_fields = value

    _headers_keys = None

    @property
    def headers_keys(self):
        return {
            'key': 'headers_keys',
            'value': self._headers_keys,
        }

    @headers_keys.setter
    def headers_keys(self, value):
        self._headers_keys = value

    _cookies_keys = None

    @property
    def cookies_keys(self):
        return {
            'key': 'cookies_keys',
            'value': self._cookies_keys,
        }

    @cookies_keys.setter
    def cookies_keys(self, value):
        self._cookies_keys = value


class BaseField:
    def __init__(self, **kwargs):
        self.typ = kwargs.get('typ', FieldType.UNKNOWN)

        self.presentment = kwargs.get('presentment', PresentmentType.OPTIONAL)

        self.min_length = kwargs.get('min_length')
        self.max_length = kwargs.get('max_length')

        self.min = kwargs.get('min')
        self.max = kwargs.get('max')

        self.error_message = kwargs.get('error_message')
        self.validator_pattern = kwargs.get('validator_pattern')

        self.description = kwargs.get('description')
        self.allowed_file_format = kwargs.get('allowed_file_format')
        self.allowed_image_format = kwargs.get('allowed_image_format')

        self.item = kwargs.get('item')
        self.value = kwargs.get('value')

    def todict(self):
        ret = {}
        if self.typ:
            ret['type'] = self.typ
        if self.presentment:
            ret['presentment'] = self.presentment
        if self.min_length:
            ret['min_length'] = self.min_length
        if self.max_length:
            ret['max_length'] = self.max_length
        if self.min:
            ret['min'] = self.min
        if self.max:
            ret['max'] = self.max
        if self.error_message:
            ret['error_message'] = self.error_message
        if self.validator_pattern:
            ret['validator_pattern'] = self.validator_pattern
        if self.description:
            ret['description'] = self.description
        if self.allowed_file_format:
            ret['allowed_file_format'] = self.allowed_file_format
        if self.allowed_image_format:
            ret['allowed_image_format'] = self.allowed_image_format
        if self.item:
            ret['item'] = self.item.todict()
        if self.value:
            ret['value'] = self.value.todict()

        return ret


class StringField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.STRING
        super().__init__(**kwargs)


class UsernameField(StringField):
    def __init__(self, **kwargs):
        kwargs['min_length'] = 3
        kwargs['max_length'] = 31
        super().__init__(**kwargs)


class PasswordField(StringField):
    def __init__(self, **kwargs):
        kwargs['min_length'] = 8
        kwargs['max_length'] = 31
        super().__init__(**kwargs)


class EmailField(StringField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class DateField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.DATE
        super().__init__(**kwargs)


class TimeField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.TIME
        super().__init__(**kwargs)


class DateTimeField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.DATETIME
        super().__init__(**kwargs)


class IntegerField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.INTEGER
        super().__init__(**kwargs)


class FloatField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.FLOAT
        super().__init__(**kwargs)


class FileField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.FILE
        super().__init__(**kwargs)


class ImageField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.IMAGE
        super().__init__(**kwargs)


class UUIDField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.UUID
        super().__init__(**kwargs)


class IDField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.ID
        super().__init__(**kwargs)


class ListField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.LIST
        super().__init__(**kwargs)


class DictionaryField(BaseField):
    def __init__(self, **kwargs):
        kwargs['typ'] = FieldType.DICTIONARY
        super().__init__(**kwargs)



from gando.architectures.services import BaseCreatorService


class BaseGetterCreatorService(BaseCreatorService):

    def __get_or_create_db_record(self):
        try:
            obj = self.model.objects.get(**self.valid_input_data)
        except self.model.DoesNotExist:
            obj = self.model.objects.create(**self.valid_input_data)
        except Exception as exc:
            raise exc
        return obj.__dict__

    def service_output_handler(self, *args, **kwargs):
        return self.__get_or_create_db_record()



from typing import List
from gando.schemas import AbstractBaseSchema


class Messenger(AbstractBaseSchema):
    type: str | None = None
    code: int | None = None
    message: str | None = None


class Data(AbstractBaseSchema):
    result: dict = dict()


class ListData(AbstractBaseSchema):
    count: int = 0
    next: str | None = None
    previous: str | None = None
    results: list = list()


class DevelopmentMessages(AbstractBaseSchema):
    info: list = list()
    warning: list = list()
    error: list = list()
    log: list = list()
    exception: list = list()


class BaseResponseSchema(AbstractBaseSchema):

    def __init__(self, response):
        response = response.json()

        success = response.get('success')
        status_code = response.get('status_code')
        has_warning = response.get('has_warning')
        exception_status = response.get('exception_status')
        monitor = response.get('monitor', {})
        messenger = response.get('messenger', [])
        many = response.get('many')
        data = response.get('data', {})
        development_messages = response.get('development_messages', {})
        kw = {
            'success': success,
            'status_code': status_code,
            'has_warning': has_warning,
            'exception_status': exception_status,
            'monitor': monitor,
            'messenger': messenger,
            'many': many,
            'data': ListData(**data) if many else Data(**data),
            'development_messages': development_messages,
        }

        super().__init__(**kw)

    success: bool = True
    status_code: int = 200
    has_warning: bool = False
    exception_status: bool = False
    monitor: dict = {}
    messenger: List[Messenger] = []
    many: bool = False
    data: Data | ListData = Data()
    development_messages: DevelopmentMessages = DevelopmentMessages()



class QueryDictSerializer:

    def __init__(self, image_fields_name=None):
        self.image_fields_name = [] if not image_fields_name else image_fields_name

    def __call__(self, input_data):
        return self.__parser(input_data)

    def __parser(self, input_data):
        if isinstance(input_data, list):
            return [self.__parser(i) for i in input_data]

        if isinstance(input_data, dict):
            tmp = dict()
            for k, v in input_data.items():
                k__split = k.split('__', 1)
                if len(k__split) == 1:
                    k_split = self.__image_field_name_parser(k)
                    if len(k_split) == 1:
                        _ = self.__parser(v)
                        v_ = _ if k not in tmp else self.__updater(tmp[k], _)
                        tmp[k] = v_ if k != 'src' else f'{self.__media_url}{v_}'
                    else:
                        _ = self.__parser({k_split[1]: v})
                        v_ = _ if k_split[0] not in tmp else self.__updater(tmp[k_split[0]], _)
                        tmp[k_split[0]] = v_ if k != 'src' else f'{self.__media_url}{v_}'
                else:
                    _ = self.__parser({k__split[1]: v})
                    v_ = _ if k__split[0] not in tmp else self.__updater(tmp[k__split[0]], _)
                    tmp[k__split[0]] = v_ if k != 'src' else f'{self.__media_url}{v_}'
            return tmp
        return input_data

    def __updater(self, a, b):
        if isinstance(a, dict) and isinstance(b, dict):
            tmp = {}
            for k_a, v_a in a.items():
                for k_b, v_b in b.items():
                    if k_a == k_b:
                        tmp[k_a] = self.__updater(v_a, v_b)
                    else:
                        tmp[k_a] = v_a
                        tmp[k_b] = v_b
            return tmp

        if isinstance(a, list):
            a.append(b)
            return a

        if isinstance(b, list):
            return [a, b]
        return [a] + [b]

    def __image_field_name_parser(self, filed_name):
        equal = True
        for img in self.image_fields_name:

            len_img = len(img)
            if len_img > len(filed_name):
                continue

            for idx in range(len_img):
                if img[idx] != filed_name[idx]:
                    equal = False
                    break
                else:
                    continue
            if equal:
                return [img] + [filed_name.split(f'{img}_', 1)[1]]

        return [filed_name]

    @property
    def __media_url(self):
        try:
            from django.conf import settings

            return settings.MEDIA_URL
        except:
            return ''



from typing import Optional, Any, Dict, List
from pydantic import BaseModel
import asyncio
import httpx
import time

GET = 'GET'
POST = 'POST'
PUT = 'PUT'
PATCH = 'PATCH'
DELETE = 'DELETE'
HEAD = 'HEAD'
OPTIONS = 'OPTIONS'
TRACE = 'TRACE'


class RequestSchema(BaseModel):
    idx: int
    url: str
    method: str = GET
    data: dict = {}
    cookies: dict = {}
    headers: dict = {}
    sch: Any


class RequestListSchema(BaseModel):
    items: List[RequestSchema] = []


class ResponseSchema(BaseModel):
    idx: int
    request: RequestSchema
    _response: Any
    status_code: int
    data: Any


class ResponseListSchema(BaseModel):
    items: List[ResponseSchema] = []
    elapsed_time: Optional[float] = 0


class AsyncoRequestManager:
    def __init__(self):
        self.__requests: RequestListSchema = RequestListSchema(items=[])
        self.__request_id_counter: int = 0
        self.__elapsed_time: float = 0
        self.__responses: ResponseListSchema = ResponseListSchema()
        self.__incomplete_requests: Dict[int, RequestSchema] = {}
        self.__incomplete_requests_temporary: Dict[int, RequestSchema] = {}

    def __add_to_incomplete_requests(self, item: RequestSchema):
        self.__incomplete_requests[item.idx] = item
        self.__incomplete_requests_temporary[item.idx] = item

    def __remove_from_incomplete_requests_temporary(self, item: RequestSchema):
        if item.idx in self.__incomplete_requests_temporary:
            del self.__incomplete_requests_temporary[item.idx]

    def __reset_incomplete_requests(self):
        self.__incomplete_requests = self.__incomplete_requests_temporary.copy()

    def add_to_requests(
        self, url, method=None, data=None, headers=None, cookies=None, schema=None):
        idx = self.__request_id_counter

        req = RequestSchema(
            idx=idx,
            url=url,
            method=method or GET,
            data=data or {},
            cookies=cookies or {},
            headers=headers or {},
            sch=schema,
        )

        self.__requests = req
        self.__add_to_incomplete_requests(req)

        self.__request_id_counter += 1
        return idx

    def __add_to_responses(self, request: RequestSchema, _response):
        idx = request.idx
        status_code = _response.status_code
        data = request.sch(**_response.json()) if request.sch else _response.json()

        rsp = ResponseSchema(
            idx=idx,
            request=request,
            _response=_response,
            status_code=status_code,
            data=data,
        )

        self.__responses.items.append(rsp)

        self.__remove_from_incomplete_requests_temporary(request)

        return idx

    async def __asyncio_request(self, client, request: RequestSchema):
        ret = False
        try:
            if request.method.upper() == GET:
                rsp = await client.get(
                    request.url,
                    cookies=request.cookies or {},
                    headers=request.headers or {}
                )
                self.__add_to_responses(request=request, _response=rsp)
                ret = True

            elif request.method.upper() == POST:
                rsp = await client.post(
                    request.url,
                    data=request.data,
                    cookies=request.cookies or {},
                    headers=request.headers or {},
                )
                self.__add_to_responses(request=request, _response=rsp)
                ret = True
            else:
                raise Exception("method not define.")
        except:
            pass
        finally:
            return ret

    def __is_completed(self):
        ret = False
        if len(self.__incomplete_requests) == 0:
            ret = True
        return ret

    async def __engine(self):
        async with httpx.AsyncClient() as client:
            await asyncio.gather(
                *[
                    self.__asyncio_request(client, i)
                    for i in self.__incomplete_requests.values()
                ]
            )
        self.__reset_incomplete_requests()

    def start(self):
        start_time = time.perf_counter()
        __stopper = 0
        while True:
            __stopper += 1
            if __stopper > 5:
                break
            if self.__is_completed():
                break
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(self.__engine())

        end_time = time.perf_counter()
        self.__responses.elapsed_time = end_time - start_time
        return self

    def export(self):
        _sort_response = self.__responses.copy()
        _sort_response.items = sorted(self.__responses.items, key=lambda item: item.idx)
        return _sort_response



from .__ import (
    BaseMessage,
    BaseFailMessage,
    BaseWarningMessage,
    BaseSuccessMessage,
    BaseInfoMessage,

    DefaultResponse100FailMessage,
    DefaultResponse200SuccessMessage,
    DefaultResponse201SuccessMessage,
    DefaultResponse300FailMessage,
    DefaultResponse400FailMessage,
    DefaultResponse401FailMessage,
    DefaultResponse403FailMessage,
    DefaultResponse404FailMessage,
    DefaultResponse421FailMessage,
    DefaultResponse500FailMessage,
)



from PIL import Image, ImageFilter
import base64
import io


def small_blur_base64(image_file_path):
    try:
        im = Image.open(image_file_path)
        (width, height) = (im.width // 8, im.height // 8)
        mimetype = im.get_format_mimetype()
        im_resized = im.resize((width, height))
        blurred_image = im_resized.filter(ImageFilter.BLUR)
        buffered = io.BytesIO()
        blurred_image.save(buffered, format="PNG")
        encoded_image = base64.b64encode(buffered.getvalue()).decode("utf-8")
        return f"data:{mimetype};base64,{encoded_image}"
    except Exception as exc:
        return



import re

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError


def validate(value):
    pattern = r"^[a-z]{1}[0-9a-z\-_]{3,31}$"
    validation_error_message = _(
        r"Please follow the format of your username correctly. "
        r"The minimum number of characters is 3 and the maximum number is 31."
        r"Please do not start your username with numbers and "
        r"symbols and do not use capital letters. "
        r"Try to use a combination of numbers and "
        r"signs (-,_) and English lowercase letters."
    )

    value = str(value)
    if not re.match(pattern, value):
        raise ValidationError(validation_error_message)
    return value



from pydantic import BaseModel as Base
from functools import lru_cache

from django.conf import settings


class ExceptionHandlerObject(Base):
    HANDLING: bool = False
    COMMUNICATION_WITH_SOFTWARE_SUPPORT: str = None


class UserAgentDeviceHandlerObject(Base):
    HANDLING: bool = False
    COOKIE_NAME: str = 'uad'


class Gando(Base):
    MONITOR_KEYS: list = list()
    DEBUG: bool = True
    CACHING: bool = False
    MONITOR: dict = dict()
    EXCEPTION_HANDLER: ExceptionHandlerObject = ExceptionHandlerObject()
    PASTE_TO_REQUEST: dict = dict()
    USER_AGENT_DEVICE_HANDLER: UserAgentDeviceHandlerObject = UserAgentDeviceHandlerObject()
    DEVELOPMENT_STATE: bool = True


@lru_cache()
def __get_settings():
    try:
        input_conf = settings.GANDO
    except:
        input_conf = {}

    try:
        input_conf['DEBUG'] = settings.DEBUG
    except:
        pass

    input_conf['EXCEPTION_HANDLER'] = (
        ExceptionHandlerObject(**input_conf['EXCEPTION_HANDLER'])
        if 'EXCEPTION_HANDLER' in input_conf
        else ExceptionHandlerObject()
    )
    input_conf['USER_AGENT_DEVICE_HANDLER'] = (
        UserAgentDeviceHandlerObject(**input_conf['USER_AGENT_DEVICE_HANDLER'])
        if 'USER_AGENT_DEVICE_HANDLER' in input_conf
        else UserAgentDeviceHandlerObject()
    )

    return Gando(**input_conf)


SETTINGS = __get_settings()



from pathlib import Path
import os
from pydantic import BaseModel as BaseSchema

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from gando.utils.strings.converters import casing
from gando.utils.strings.casings import PASCAL_CASE, KEBAB_CASE

# definitions
PROJECT_PATH = settings.BASE_DIR


# tools
def package_maker(parent_path: str, name: str) -> str:
    package_path = os.path.join(parent_path, casing(name))
    if not os.path.exists(package_path):
        os.mkdir(package_path)

    package_init = os.path.join(package_path, '__init__.py')
    if not os.path.exists(package_init):
        Path(package_init).touch()

    return package_path


def python_file_maker(path: str, name: str, private_=False) -> str:
    name = casing(f'{"__" if private_ else ""}{name.split(".")[0]}')
    python_file_full_name = os.path.join(path, f'{name}.py')
    if not os.path.exists(python_file_full_name):
        Path(python_file_full_name).touch()

    return python_file_full_name


class BasePackageInfo:
    def __init__(self, path: str):
        self.path: str = path
        self.initial_path: str = os.path.join(self.path, '__init__.py')


class BasePackages:
    def __init__(self, application_path: str):
        self.repo: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=application_path, name='repo'))
        self.repo__schemas: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='schemas'))
        self.repo__schemas__interfaces: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo__schemas.path, name='interfaces'))
        self.repo__interfaces: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='interfaces'))


class Command(BaseCommand):
    help = ("This command helps us to automatically create the basic prerequisites of "
            "this interface when we define a interface, so that we can personalize each of "
            "the tools that are created automatically if needed.")

    def add_arguments(self, parser):
        """
        This command has two mandatory arguments so that we can automatically create
        the prerequisites for each interface.
        -al or --applabel: This argument is used to get the name of the desired application.
        -in or --interfacename: And this argument also specifies the name of the interface inside that application.
        """
        parser.add_argument(
            '-al', '--applabel', type=str,
            help='This argument is used to get the name of the desired application.')
        parser.add_argument(
            '-in', '--interfacename', type=str,
            help='This argument also specifies the name of the interface inside that application.')

    def handle(self, *args, **kwargs):
        # First, we get the value of the set parameters from the sent command.
        self.app_label = kwargs
        self.interface_name = kwargs

        # Here we check the acceptability of the received parameters.
        self.interface_name_is_valid(rais_exception=True)

        # set Application path
        self.application_path = kwargs

        # set Base Packages path
        self.base_packages = BasePackages(application_path=self.application_path)

        self.initial_interface()

        self.initial_schemas()

    __interface_name: str | None = None

    @property
    def interface_name(self) -> str | None:
        return self.__interface_name

    @interface_name.setter
    def interface_name(self, value: dict):
        tmp = value.get('interfacename', None)
        if tmp:
            self.__interface_name = casing(tmp, to_case=PASCAL_CASE)

    @property
    def interface_name_snake_case(self):
        return casing(self.interface_name)

    __app_label: str | None = None

    @property
    def app_label(self) -> str | None:
        return self.__app_label

    @app_label.setter
    def app_label(self, value: dict):
        self.__app_label = value.get('applabel', None)

    def interface_name_is_valid(self, rais_exception=True) -> bool:
        if not self.app_label:
            if rais_exception:
                raise CommandError(
                    "The name of the application is not entered, "
                    "which is actually the same as the -al or --applabel argument. "
                    "It is not set. Please check and fix the error and try again.")
            return False

        if not self.interface_name:
            if rais_exception:
                raise CommandError(
                    "The interface name is not entered, "
                    "which is actually the same as the -mn or --interfacename argument. "
                    "It is not set. Please check and fix the error and try again.")
            return False

        return True

    def interface_is_exist(self) -> bool:
        try:
            from django.apps import apps

            apps.get_interface(app_label=self.app_label, interface_name=self.interface_name)
            return True
        except:
            return False

    __application_path: str | None = None

    @property
    def application_path(self) -> str | None:
        return self.__application_path

    @application_path.setter
    def application_path(self, value: dict):
        self.__application_path = os.path.join(PROJECT_PATH, value.get('applabel'))

    base_packages: BasePackages | None = None

    def initial_interface(self):
        file_path = python_file_maker(self.base_packages.repo__interfaces.path, self.interface_name, private_=False)

        with open(file_path, 'w') as f:
            f.write(
                f"from gando.architectures.interfaces import BaseInterface"
                f"\n"
                f"\n"
                f"\n"
                f"class {self.interface_name}Interface(BaseInterface):"
                f"\n"
                f"    def __init__(self, *args, **kwargs):"
                f"\n"
                f"        super().__init__(*args, **kwargs)"
                f"\n"
            )

    def initial_schemas(self):
        p_path = package_maker(
            self.base_packages.repo__schemas__interfaces.path, self.interface_name_snake_case)
        python_file_maker(p_path, 'input_data', private_=True)
        python_file_maker(p_path, 'output_data', private_=True)
        python_file_maker(p_path, 'params', private_=True)

    @staticmethod
    def __new_line(file_name):
        tmp = ''
        with open(file_name, 'r') as f:
            lines = f.read()
            if lines[-1] != '\n':
                tmp = '\n'

        ret = tmp
        return ret

    @staticmethod
    def __find_in_file(file_name, sub):
        tmp = ''
        with open(file_name, 'r') as f:
            txt = f.read()
            if txt.find(sub) == -1:
                tmp = None

        ret = tmp
        return ret

    @staticmethod
    def __file_content(file_name, default=''):
        tmp = default
        with open(file_name, 'r') as f:
            txt = f.read()
            if len(txt) != 0:
                tmp = str(txt)

        ret = tmp
        return ret



def logged_in(*args, **kwargs):
    request = kwargs.get('request')
    if not request:
        return False

    return request.user.is_authenticated


def user_first_name(*args, **kwargs):
    request = kwargs.get('request')
    if not request or not hasattr(request, 'user') or not hasattr(request.user, 'first_name'):
        return None

    return request.user.first_name


def user_last_name(*args, **kwargs):
    request = kwargs.get('request')
    if not request or not hasattr(request, 'user') or not hasattr(request.user, 'last_name'):
        return None

    return request.user.last_name


def user_username(*args, **kwargs):
    request = kwargs.get('request')
    if not request or not hasattr(request, 'user') or not hasattr(request.user, 'username'):
        return None

    return request.user.username



from PIL import Image as ImagePIL, ImageFilter
import base64
import io


class Image:
    def __init__(self, src):
        self.__src = src

        self._width = None
        self._height = None

        self._size = None

        self._proportion = None

        self._file_format = None

        self._b64 = None

        self._open_image = None

    def get_width(self):
        if not self._width:
            self._width = self.__open_image().width
        return self._width

    def get_height(self):
        if not self._height:
            self._height = self.__open_image().height
        return self._height

    def get_size(self):
        if not self._size:
            self._size = self.__src.size
        return self._size

    def get_proportion(self):
        if not self._proportion:
            self._proportion = self.__proportion()
        return self._proportion

    def get_file_format(self):
        if not self._file_format:
            self._file_format = self.__src.name.split('.')[-1].lower()
        return self._file_format

    def get_b64(self):
        if not self._b64:
            self._b64 = self.__b64()
        return self._b64

    def __open_image(self):
        if not self._open_image:
            self._open_image = ImagePIL.open(self.__src)
        return self._open_image

    def __b64(self):
        try:
            (b64width, b64height) = (self.get_width() // 8, self.get_height() // 8)

            mimetype = self.__open_image().get_format_mimetype()
            im_resized = self.__open_image().resize((b64width, b64height))
            blurred_image = im_resized.filter(ImageFilter.BLUR)
            buffered = io.BytesIO()
            blurred_image.save(buffered, format='PNG')
            encoded_image = base64.b64encode(buffered.getvalue()).decode('utf-8')

            return f'data:{mimetype};base64,{encoded_image}'
        except:
            return ''

    def __proportion(self) -> int:
        image_p = round(self.get_height() / self.get_width(), 15)

        b__p_9_16__p_2_3 = round(9 / 16 + abs(9 / 16 - 2 / 3) / 2, 15)
        b__p_2_3__p_3_4 = round(2 / 3 + abs(2 / 3 - 3 / 4) / 2, 15)
        b__p_3_4__p_4_5 = round(3 / 4 + abs(3 / 4 - 4 / 5) / 2, 15)
        b__p_4_5__p_1_1 = round(4 / 5 + abs(4 / 5 - 1 / 1) / 2, 15)
        b__p_1_1__p_5_4 = round(1 / 1 + abs(1 / 1 - 5 / 4) / 2, 15)
        b__p_5_4__p_4_3 = round(5 / 4 + abs(5 / 4 - 4 / 3) / 2, 15)
        b__p_4_3__p_3_2 = round(4 / 3 + abs(4 / 3 - 3 / 2) / 2, 15)
        b__p_3_2__p_16_9 = round(3 / 2 + abs(3 / 2 - 16 / 9) / 2, 15)

        if image_p <= b__p_9_16__p_2_3:
            return 916

        elif b__p_9_16__p_2_3 < image_p <= b__p_2_3__p_3_4:
            return 23

        elif b__p_2_3__p_3_4 < image_p <= b__p_3_4__p_4_5:
            return 34

        elif b__p_3_4__p_4_5 < image_p <= b__p_4_5__p_1_1:
            return 45

        elif b__p_4_5__p_1_1 < image_p <= b__p_1_1__p_5_4:
            return 11

        elif b__p_1_1__p_5_4 < image_p <= b__p_5_4__p_4_3:
            return 54

        elif b__p_5_4__p_4_3 < image_p <= b__p_4_3__p_3_2:
            return 43

        elif b__p_4_3__p_3_2 < image_p <= b__p_3_2__p_16_9:
            return 32

        else:
            return 169



from django.contrib import admin

from gando.models import UserAgentDevice


@admin.register(UserAgentDevice)
class UserAgentDeviceModelAdmin(admin.ModelAdmin):
    list_display = (
        'id',

        'key',

        'user',

        'created_dt',

        'ip',

        'user_agent_is_mobile',
        'user_agent_is_tablet',
        'user_agent_is_touch_capable',
        'user_agent_is_pc',
        'user_agent_is_bot',

        'user_agent_browser_family',
        'user_agent_browser_version',

        'user_agent_os_family',
        'user_agent_os_version',

        'user_agent_device_family',
        'user_agent_device_brand',
        'user_agent_device_model',
    )
    list_display_links = (
        'id',

        'key',

        'user',

        'created_dt',

        'ip',

        'user_agent_is_mobile',
        'user_agent_is_tablet',
        'user_agent_is_touch_capable',
        'user_agent_is_pc',
        'user_agent_is_bot',

        'user_agent_browser_family',
        'user_agent_browser_version',

        'user_agent_os_family',
        'user_agent_os_version',

        'user_agent_device_family',
        'user_agent_device_brand',
        'user_agent_device_model',
    )
    search_fields = (
        'id',
        'user__id',
        'ip',

        'key',
    )
    list_filter = (
        'user_agent_is_mobile',
        'user_agent_is_tablet',
        'user_agent_is_touch_capable',
        'user_agent_is_pc',
        'user_agent_is_bot',

        'user_agent_browser_family',
        'user_agent_browser_version',

        'user_agent_os_family',
        'user_agent_os_version',

        'user_agent_device_family',
        'user_agent_device_brand',
        'user_agent_device_model',
    )
    readonly_fields = ('id', 'created_dt', 'key',)
    fieldsets = [
        ('Base Info', {'fields': [
            'id',
            'user',
            'key',
            'ip',
            'created_dt',
        ]}),
        ('UserAgent', {'fields': [
            'user_agent_is_mobile',
            'user_agent_is_tablet',
            'user_agent_is_touch_capable',
            'user_agent_is_pc',
            'user_agent_is_bot',

            'user_agent_browser_family',
            'user_agent_browser_version',

            'user_agent_os_family',
            'user_agent_os_version',

            'user_agent_device_family',
            'user_agent_device_brand',
            'user_agent_device_model',
        ]}),
    ]

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False



from pydantic import BaseModel
from datetime import datetime


class AbstractBaseSchema(BaseModel):
    pass


class AbstractBaseSchemaModel(AbstractBaseSchema):
    id: int | None = None
    uid: str | None = None
    created_dt: datetime | None = None
    updated_dt: datetime | None = None
    available: int | None = None



import re

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError


def validate(value):
    pattern = r"^[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?$"
    validation_error_message = _(
        "Please enter your email correctly. The correct email format is as follows."
        "example@example@example")

    value = str(value)
    if not re.match(pattern, value):
        raise ValidationError(validation_error_message)
    return value



class InfoStringMessage(str):
    code = 'info'

    def __new__(cls, string, code=None):
        self = super().__new__(cls, string)
        self.code = code
        return self

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        try:
            return result and self.code == other.code
        except AttributeError:
            return result

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __repr__(self):
        return 'InfoStringMessage(string=%r, code=%r)' % (
            str(self),
            self.code,
        )

    def __hash__(self):
        return hash(str(self))


class ErrorStringMessage(str):
    code = 'error'

    def __new__(cls, string, code=None):
        self = super().__new__(cls, string)
        self.code = code
        return self

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        try:
            return result and self.code == other.code
        except AttributeError:
            return result

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __repr__(self):
        return 'ErrorStringMessage(string=%r, code=%r)' % (
            str(self),
            self.code,
        )

    def __hash__(self):
        return hash(str(self))


class WarningStringMessage(str):
    code = 'warning'

    def __new__(cls, string, code=None):
        self = super().__new__(cls, string)
        self.code = code
        return self

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        try:
            return result and self.code == other.code
        except AttributeError:
            return result

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __repr__(self):
        return 'WarningStringMessage(string=%r, code=%r)' % (
            str(self),
            self.code,
        )

    def __hash__(self):
        return hash(str(self))


class LogStringMessage(str):
    code = 'log'

    def __new__(cls, string, code=None):
        self = super().__new__(cls, string)
        self.code = code
        return self

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        try:
            return result and self.code == other.code
        except AttributeError:
            return result

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __repr__(self):
        return 'LogStringMessage(string=%r, code=%r)' % (
            str(self),
            self.code,
        )

    def __hash__(self):
        return hash(str(self))


class ExceptionStringMessage(str):
    code = 'exception'

    def __new__(cls, string, code=None):
        self = super().__new__(cls, string)
        self.code = code
        return self

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        try:
            return result and self.code == other.code
        except AttributeError:
            return result

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __repr__(self):
        return 'ExceptionStringMessage(string=%r, code=%r)' % (
            str(self),
            self.code,
        )

    def __hash__(self):
        return hash(str(self))



from pathlib import Path
import os
from pydantic import BaseModel as BaseSchema

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from gando.utils.strings.converters import casing
from gando.utils.strings.casings import PASCAL_CASE, KEBAB_CASE

# definitions
PROJECT_PATH = settings.BASE_DIR


# tools
def package_maker(parent_path: str, name: str) -> str:
    package_path = os.path.join(parent_path, casing(name))
    if not os.path.exists(package_path):
        os.mkdir(package_path)

    package_init = os.path.join(package_path, '__init__.py')
    if not os.path.exists(package_init):
        Path(package_init).touch()

    return package_path


def python_file_maker(path: str, name: str, private_=False) -> str:
    name = casing(f'{"__" if private_ else ""}{name.split(".")[0]}')
    python_file_full_name = os.path.join(path, f'{name}.py')
    if not os.path.exists(python_file_full_name):
        Path(python_file_full_name).touch()

    return python_file_full_name


class BaseFiles:
    def __init__(self, application_path: str):
        self.admin: str = python_file_maker(path=application_path, name='admin')
        self.models: str = python_file_maker(path=application_path, name='models')
        self.urls: str = python_file_maker(path=application_path, name='urls')


class BasePackageInfo:
    def __init__(self, path: str):
        self.path: str = path
        self.initial_path: str = os.path.join(self.path, '__init__.py')


class BasePackages:
    def __init__(self, application_path: str):
        self.repo: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=application_path, name='repo'))
        self.repo__admin: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='admin'))
        self.repo__models: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='models'))
        self.repo__schemas: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='schemas'))
        self.repo__schemas__models: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo__schemas.path, name='models'))
        self.repo__schemas__apis: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo__schemas.path, name='apis'))
        self.repo__schemas__services: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo__schemas.path, name='services'))
        self.repo__urls: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='urls'))
        self.repo__services: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='services'))
        self.repo__interfaces: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='interfaces'))
        self.repo__apis: BasePackageInfo = BasePackageInfo(
            path=package_maker(parent_path=self.repo.path, name='apis'))


class Command(BaseCommand):
    help = ("This command helps us to automatically create the basic prerequisites of "
            "this model when we define a model, so that we can personalize each of "
            "the tools that are created automatically if needed.")

    def add_arguments(self, parser):
        """
        This command has two mandatory arguments so that we can automatically create
        the prerequisites for each model.
        -al or --applabel: This argument is used to get the name of the desired application.
        -mn or --modelname: And this argument also specifies the name of the model inside that application.
        """
        parser.add_argument(
            '-al', '--applabel', type=str,
            help='This argument is used to get the name of the desired application.')
        parser.add_argument(
            '-mn', '--modelname', type=str,
            help='This argument also specifies the name of the model inside that application.')

    def handle(self, *args, **kwargs):
        # First, we get the value of the set parameters from the sent command.
        self.app_label = kwargs
        self.model_name = kwargs

        # Here we check the acceptability of the received parameters.
        self.model_name_is_valid(rais_exception=True)

        # set Application path
        self.application_path = kwargs

        # set Base Files path
        self.base_files = BaseFiles(application_path=self.application_path)
        # set Base Packages path
        self.base_packages = BasePackages(application_path=self.application_path)

        self.initial_model()

        self.initial_admin()
        self.initial_urlpatterns()
        self.initial_schemas()

    __model_name: str | None = None

    @property
    def model_name(self) -> str | None:
        return self.__model_name

    @model_name.setter
    def model_name(self, value: dict):
        tmp = value.get('modelname', None)
        if tmp:
            self.__model_name = casing(tmp, to_case=PASCAL_CASE)

    @property
    def model_name_snake_case(self):
        return casing(self.model_name)

    __app_label: str | None = None

    @property
    def app_label(self) -> str | None:
        return self.__app_label

    @app_label.setter
    def app_label(self, value: dict):
        self.__app_label = value.get('applabel', None)

    def model_name_is_valid(self, rais_exception=True) -> bool:
        if not self.app_label:
            if rais_exception:
                raise CommandError(
                    "The name of the application is not entered, "
                    "which is actually the same as the -al or --applabel argument. "
                    "It is not set. Please check and fix the error and try again.")
            return False

        if not self.model_name:
            if rais_exception:
                raise CommandError(
                    "The model name is not entered, "
                    "which is actually the same as the -mn or --modelname argument. "
                    "It is not set. Please check and fix the error and try again.")
            return False

        return True

    def model_is_exist(self) -> bool:
        try:
            from django.apps import apps

            apps.get_model(app_label=self.app_label, model_name=self.model_name)
            return True
        except:
            return False

    __application_path: str | None = None

    @property
    def application_path(self) -> str | None:
        return self.__application_path

    @application_path.setter
    def application_path(self, value: dict):
        self.__application_path = os.path.join(PROJECT_PATH, value.get('applabel'))

    base_files: BaseFiles | None = None
    base_packages: BasePackages | None = None

    def initial_model(self):
        file_path = python_file_maker(self.base_packages.repo__models.path, self.model_name, private_=True)

        with open(file_path, 'w') as f:
            f.write(f'from django.db import models\n\n\nclass {self.model_name}(models.Model):\n    pass\n')

        i_fp = self.base_packages.repo__models.initial_path
        with open(i_fp, 'a') as f:
            f.write(
                f"{self.__new_line(i_fp)}"
                f"from .__{self.model_name_snake_case} import {self.model_name} as {self.model_name}Model"
                f"\n"
            )

        m_fp = self.base_files.models
        with open(m_fp, 'a') as f:
            f.write(
                f"{self.__new_line(m_fp)}"
                f"from .repo.models import {self.model_name}Model"
                f"\n"
            )

    def initial_admin(self):
        file_path = python_file_maker(self.base_packages.repo__admin.path, self.model_name, private_=True)

        with open(file_path, 'w') as f:
            f.write(
                f"from django.contrib import admin\n\n"
                f"from {self.app_label}.models import {self.model_name}Model as Model\n\n\n"
                f"@admin.register(Model)\n"
                f"class {self.model_name}Admin(admin.ModelAdmin):\n    pass\n"
            )

        i_fp = self.base_packages.repo__admin.initial_path
        with open(i_fp, 'a') as f:
            f.write(
                f"{self.__new_line(i_fp)}"
                f"from .__{self.model_name_snake_case} import {self.model_name}Admin"
                f"\n"
            )

        a_fp = self.base_files.admin
        with open(a_fp, 'a') as f:
            f.write(
                f"{self.__new_line(a_fp)}"
                f"from .repo.admin import {self.model_name}Admin"
                f"\n"
            )

    def initial_urlpatterns(self):
        file_path = python_file_maker(self.base_packages.repo__urls.path, self.model_name)

        with open(file_path, 'w') as f:
            f.write(
                f"from django.urls import path\n\n"
                f"app_name = '{self.model_name_snake_case}s'\n"
                f"urlpatterns = [\n"
                f"    # path('', SampleAPI.as_view(), name='sample'),\n"
                f"]\n"
            )

        default_txt = (
            f"from django.urls import path, include\n\n\n"
            f"app_name = '{self.app_label}'\n"
            f"urlpatterns = [\n"
            f"]\n"
        )
        u_fp = self.base_files.urls
        if self.__find_in_file(u_fp, 'urlpatterns') is not None:
            txt = self.__file_content(u_fp)
        else:
            txt = default_txt
        with open(self.base_files.urls, 'w') as f:
            f.write(
                f"{txt[:txt.find(']')]}"
                f"\n"
                f"    path('{casing(self.model_name, to_case=KEBAB_CASE)}s/', "
                f"include('{self.app_label}.repo.urls.{self.model_name_snake_case}')),"
                f"\n"
                f"{txt[txt.find(']'):]}"
            )

    def initial_schemas(self):
        file_path = python_file_maker(self.base_packages.repo__schemas__models.path, self.model_name, private_=True)
        with open(file_path, 'w') as f:
            f.write(
                f"from pydantic import BaseModel\n\n\n"
                f"class {self.model_name}ModelSchema(BaseModel):\n"
                f"    id: int | None = None\n"
            )

        i_fp = self.base_packages.repo__schemas__models.initial_path
        with open(i_fp, 'a') as f:
            f.write(
                f"{self.__new_line(i_fp)}"
                f"from .__{self.model_name_snake_case} import {self.model_name}ModelSchema"
                f"\n"
            )

    @staticmethod
    def __new_line(file_name):
        tmp = ''
        with open(file_name, 'r') as f:
            lines = f.read()
            if lines[-1] != '\n':
                tmp = '\n'

        ret = tmp
        return ret

    @staticmethod
    def __find_in_file(file_name, sub):
        tmp = ''
        with open(file_name, 'r') as f:
            txt = f.read()
            if txt.find(sub) == -1:
                tmp = None

        ret = tmp
        return ret

    @staticmethod
    def __file_content(file_name, default=''):
        tmp = default
        with open(file_name, 'r') as f:
            txt = f.read()
            if len(txt) != 0:
                tmp = str(txt)

        ret = tmp
        return ret



import re

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

from datetime import datetime


def validate(value):
    pattern = r'^(18[0-9]{2}|19[0-9]{2}|2[0-9]{3})-(0[1-9]|1[012])-([123]0|[012][1-9]|31) ([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9])$'
    validation_error_message = _(
        "The entered value must be greater than '1800-01-01 00:00:00' and "
        "entered in the following format.\nYYYY-MM-DD hh:mm:ss")

    value = str(value)
    if not re.match(pattern, value):
        raise ValidationError(validation_error_message)

    value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
    return value



from abc import abstractmethod
from types import NoneType
from pydantic import BaseModel

from django.core.cache import cache as djcache

from gando.config import SETTINGS


class CacheConfig(BaseModel):
    cache_prefix_key: str | None = None
    cache_suffix_key: str | None = None
    cache_middle_separator_key: str | None = None
    cache: bool | None = None
    cache_timeout: int | None = None
    cache_unique_key: str | tuple | list | set | None = None


class BaseService:
    """
    - Config:
        - type:
            - class

        - attrs:
            - cache:
                - type:
                    - bool
                - descriptions:
                    - Using this variable,
                    you can activate or deactivate the service's automatic caching system.
                    The default mode is disabled.

            - cache_key:
                - type:
                    - str or tuple or list or set
                - descriptions:
                    - You can choose a variable or a combination of a number of variables as
                    a unique key for caching.
                    If there is no such key,
                    the same name is selected as the unique key.

            - cache_prefix_key:
                - type:
                    - str

            - cache_middle_separator_key:
                - type:
                    - str

            - cache_suffix_key:
                - type:
                    - str

            - cache_timeout:
                - type:
                    - int or None
    """

    def __init__(self, *args, **kwargs):
        self.__input_cache_config: CacheConfig = CacheConfig(**kwargs)

    def execute(self, reset_cache=False, *args, **kwargs):
        if SETTINGS.CACHING and self.cache:
            if reset_cache:
                output = self.service_output_handler(*args, **kwargs)
            else:
                tmp = djcache.get(key=self.cache_unique_key())
                if tmp is None:
                    tmp = self.service_output_handler(self, *args, **kwargs)
                output = tmp
            djcache.set(key=self.cache_unique_key(), value=output)
        else:
            output = self.service_output_handler(self, *args, **kwargs)

        ret = output
        return ret

    @abstractmethod
    def service_output_handler(self, *args, **kwargs):
        pass

    @property
    def cache_prefix_key(self):
        if self.__input_cache_config.cache_prefix_key is not None:
            return self.__input_cache_config.cache_prefix_key

        if (
            hasattr(self, 'Config') and
            hasattr(self.Config, 'cache_prefix_key') and
            isinstance(self.Config.cache_prefix_key, str)
        ):
            return self.Config.cache_prefix_key

        return ''

    @property
    def cache_suffix_key(self):
        if self.__input_cache_config.cache_suffix_key is not None:
            return self.__input_cache_config.cache_suffix_key

        if (
            hasattr(self, 'Config') and
            hasattr(self.Config, 'cache_suffix_key') and
            isinstance(self.Config.cache_suffix_key, str)
        ):
            return self.Config.cache_suffix_key

        return ''

    @property
    def cache_middle_separator_key(self):
        if self.__input_cache_config.cache_middle_separator_key is not None:
            return self.__input_cache_config.cache_middle_separator_key

        if (
            hasattr(self, 'Config') and
            hasattr(self.Config, 'cache_middle_separator_key') and
            isinstance(self.Config.cache_middle_separator_key, str)
        ):
            return self.Config.cache_middle_separator_key

        return ''

    @property
    def cache(self):
        if self.__input_cache_config.cache is not None:
            return self.__input_cache_config.cache

        if (
            hasattr(self, 'Config') and
            hasattr(self.Config, 'cache') and
            isinstance(self.Config.cache, bool)
        ):
            return self.Config.cache

        return False

    @property
    def cache_timeout(self):
        if self.__input_cache_config.cache_timeout is not None:
            return self.__input_cache_config.cache_timeout

        ret = 5
        if hasattr(self, 'Config') and hasattr(self.Config, 'cache_timeout'):
            if isinstance(self.Config.cache_timeout, int) or isinstance(self.Config.cache_timeout, NoneType):
                ret = self.Config.cache_timeout

        return ret

    def cache_unique_key(self):
        if self.__input_cache_config.cache_unique_key is not None:
            return self.__input_cache_config.cache_unique_key

        if hasattr(self, 'Config') and hasattr(self.Config, 'cache_key'):
            keys = []
            if isinstance(self.Config.cache_key, str):
                keys = [self.Config.cache_key]

            elif (
                isinstance(self.Config.cache_key, tuple) or
                isinstance(self.Config.cache_key, list) or
                isinstance(self.Config.cache_key, set)
            ):
                keys = [i for i in self.Config.cache_key]

            else:
                pass

            ret = (
                self.cache_prefix_key +
                self.cache_middle_separator_key.join(
                    str(
                        getattr(self, i) if hasattr(self, i) else i.upper()
                    ) for i in keys
                ) +
                self.cache_suffix_key
            )
            return ret

        return ''



from pydantic import BaseModel


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


class UADSchema(BaseModel):
    def __init__(self, request):
        kw = {}
        if hasattr(request, 'user_agent'):
            user_agent = request.user_agent
            kw = {
                'user_id': request.user.id if request.user.is_authenticated else None,

                'user_agent_is_mobile': user_agent.is_mobile,
                'user_agent_is_tablet': user_agent.is_tablet,
                'user_agent_is_touch_capable': user_agent.is_touch_capable,
                'user_agent_is_pc': user_agent.is_pc,
                'user_agent_is_bot': user_agent.is_bot,

                'user_agent_browser_family': user_agent.browser.family,
                'user_agent_browser_version': user_agent.browser.version_string,

                'user_agent_os_family': user_agent.os.family,
                'user_agent_os_version': user_agent.os.version_string,

                'user_agent_device_family': user_agent.device.family,
                'user_agent_device_brand': user_agent.device.brand,
                'user_agent_device_model': user_agent.device.model,

                'ip': get_client_ip(request)
            }
        super().__init__(**kw)

    user_id: int | None = None
    user_agent_is_mobile: bool | None = None
    user_agent_is_tablet: bool | None = None
    user_agent_is_touch_capable: bool | None = None
    user_agent_is_pc: bool | None = None
    user_agent_is_bot: bool | None = None
    user_agent_browser_family: str | None = None
    user_agent_browser_version: str | None = None
    user_agent_os_family: str | None = None
    user_agent_os_version: str | None = None
    user_agent_device_family: str | None = None
    user_agent_device_brand: str | None = None
    user_agent_device_model: str | None = None
    ip: str | None = None



from django.contrib.auth.hashers import get_hasher, identify_hasher


def make_hash_string(value, salt=None, hasher='default'):
    if value is None:
        return None

    value = str(value)
    hasher = get_hasher(hasher)
    salt = salt or hasher.salt()

    ret = hasher.encode(value, salt)
    return ret


def check_hash_string(value, encoded, setter=None, preferred='default'):
    if not value:
        return False

    preferred = get_hasher(preferred)
    try:
        hasher = identify_hasher(encoded)
    except ValueError:
        return False

    hasher_changed = hasher.algorithm != preferred.algorithm
    must_update = hasher_changed or preferred.must_update(encoded)
    is_correct = hasher.verify(value, encoded)

    if not is_correct and not hasher_changed and must_update:
        hasher.harden_runtime(value, encoded)

    if setter and is_correct and must_update:
        setter(value)

    ret = is_correct
    return ret



from simple_history.models import HistoricalRecords
import uuid

from django.utils.translation import gettext_lazy as _
from django.db import models
from django.db.models import Manager as DjBaseManager, QuerySet


class BaseManager(DjBaseManager):
    def get_queryset(self):
        queryset = super().get_queryset()
        queryset = queryset.filter(available=1)
        return queryset


class Manager(BaseManager.from_queryset(QuerySet)):
    pass


class AbstractBaseModel(models.Model):
    id = models.UUIDField(
        verbose_name=_('ID'),
        primary_key=True,
        default=uuid.uuid4,
        blank=False,
        null=False,
        unique=True,
        editable=False,
        db_index=True,
    )
    created_dt = models.DateTimeField(
        verbose_name=_('Created Datetime'),
        auto_now_add=True,
        db_index=True,
    )
    updated_dt = models.DateTimeField(
        verbose_name=_('Updated Datetime'),
        auto_now=True,
        db_index=True,
    )
    available = models.IntegerField(
        verbose_name=_('Available'),
        default=1,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
        db_index=True,
    )
    history = HistoricalRecords(inherit=True)

    objects = Manager()

    class Meta:
        abstract = True


class WhiteAbstractBaseModel(models.Model):
    id = models.UUIDField(
        verbose_name=_('ID'),
        primary_key=True,
        default=uuid.uuid4,
        blank=False,
        null=False,
        unique=True,
        editable=False,
        db_index=True,
    )
    created_dt = models.DateTimeField(
        verbose_name=_('Created Datetime'),
        auto_now_add=True,
        db_index=True,
    )
    updated_dt = models.DateTimeField(
        verbose_name=_('Updated Datetime'),
        auto_now=True,
        db_index=True,
    )
    available = models.IntegerField(
        verbose_name=_('Available'),
        default=1,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
        db_index=True,
    )
    history = HistoricalRecords(inherit=True)

    class Meta:
        abstract = True


class AbstractBaseModelFaster(models.Model):
    created_dt = models.DateTimeField(
        verbose_name=_('Created Datetime'),
        auto_now_add=True,
        db_index=True,
    )
    updated_dt = models.DateTimeField(
        verbose_name=_('Updated Datetime'),
        auto_now=True,
        db_index=True,
    )
    available = models.IntegerField(
        verbose_name=_('Available'),
        default=1,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
        db_index=True,
    )

    history = HistoricalRecords(inherit=True)

    class Meta:
        abstract = True



import importlib
from inspect import currentframe, getframeinfo
from typing import Any

from django.core.exceptions import PermissionDenied
from django.db import connections
from django.http import Http404
from pydantic import BaseModel
from rest_framework import exceptions
from rest_framework import status
from rest_framework.exceptions import ErrorDetail
from rest_framework.generics import (
    CreateAPIView as DRFGCreateAPIView,
    ListAPIView as DRFGListAPIView,
    RetrieveAPIView as DRFGRetrieveAPIView,
    UpdateAPIView as DRFGUpdateAPIView,
    DestroyAPIView as DRFGDestroyAPIView,
)
from rest_framework.response import Response
from rest_framework.views import APIView

from gando.config import SETTINGS
from gando.http.api_exceptions import (
    EnduserResponseAPIMessage,
    DeveloperResponseAPIMessage,

    DeveloperExceptionResponseAPIMessage,
    DeveloperErrorResponseAPIMessage,
    DeveloperWarningResponseAPIMessage,
    EnduserFailResponseAPIMessage,
    EnduserErrorResponseAPIMessage,
    EnduserWarningResponseAPIMessage,
)
from gando.http.responses.string_messages import (
    InfoStringMessage,
    ErrorStringMessage,
    WarningStringMessage,
    LogStringMessage,
    ExceptionStringMessage,
)
from gando.utils.exceptions import PassException
from gando.utils.messages import (
    DefaultResponse100FailMessage,
    DefaultResponse200SuccessMessage,
    DefaultResponse201SuccessMessage,
    DefaultResponse300FailMessage,
    DefaultResponse400FailMessage,
    DefaultResponse401FailMessage,
    DefaultResponse403FailMessage,
    DefaultResponse404FailMessage,
    DefaultResponse421FailMessage,
    DefaultResponse500FailMessage,
)


def _valid_user(user_id, request):
    from django.contrib.auth import get_user_model

    try:
        obj = get_user_model().objects.get(id=request.user.id)
        obj_id = obj.id if isinstance(obj.id, int) else str(obj.id)
        if obj_id == user_id:
            return True
        return False
    except Exception as exc:
        return False


def set_rollback():
    for db in connections.all():
        if db.settings_dict['ATOMIC_REQUESTS'] and db.in_atomic_block:
            db.set_rollback(True)


class BaseAPI(APIView):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__messenger = []

        self.__data = None

        self.__logs_message = []
        self.__infos_message = []
        self.__warnings_message = []
        self.__errors_message = []
        self.__exceptions_message = []

        self.__monitor: dict = dict()

        self.__status_code: int | None = None

        self.__headers: dict = dict()
        self.__cookies_for_set: list = list()
        self.__cookies_for_delete: list = list()

        self.__content_type: str | None = None
        self.__exception_status: bool = False

    def __paste_to_request_func_loader(self, f, request, *args, **kwargs):
        try:
            mod_name, func_name = f.rsplit('.', 1)
            mod = importlib.import_module(mod_name)
            func = getattr(mod, func_name)

            return func(request=request, *args, **kwargs)
        except PassException as exc:
            frame_info = getframeinfo(currentframe())
            self.set_log_message(
                key='pass',
                value=f"message:{exc.args[0]}, "
                      f"file_name: {frame_info.filename}, "
                      f"line_number: {frame_info.lineno}")
            return None

    def paste_to_request_func_loader_play(self, request, *args, **kwargs):
        for key, f in SETTINGS.PASTE_TO_REQUEST.items():
            rslt = self.__paste_to_request_func_loader(f, request, *args, **kwargs)
            if rslt:
                setattr(request, key, rslt)

        return request

    def initialize_request(self, request, *args, **kwargs):

        request_ = super().initialize_request(request, *args, **kwargs)
        rslt = self.paste_to_request_func_loader_play(request_)
        ret = rslt
        return ret

    def handle_exception(self, exc):

        if isinstance(exc, DeveloperResponseAPIMessage):
            if isinstance(exc, DeveloperErrorResponseAPIMessage):
                self.set_status_code(exc.status_code)
                self.set_error_message(key=exc.code, value=exc.message)

            elif isinstance(exc, DeveloperExceptionResponseAPIMessage):
                self.set_status_code(exc.status_code)
                self.set_exception_message(key=exc.code, value=exc.message)

            elif isinstance(exc, DeveloperWarningResponseAPIMessage):
                self.set_status_code(exc.status_code)
                self.set_warning_message(key=exc.code, value=exc.message)

        if isinstance(exc, EnduserResponseAPIMessage):
            if isinstance(exc, EnduserErrorResponseAPIMessage):
                self.set_status_code(exc.status_code)
                self.add_error_message_to_messenger(code=exc.code, message=exc.message)

            elif isinstance(exc, EnduserFailResponseAPIMessage):
                self.set_status_code(exc.status_code)
                self.add_fail_message_to_messenger(code=exc.code, message=exc.message)

            elif isinstance(exc, EnduserWarningResponseAPIMessage):
                self.set_status_code(exc.status_code)
                self.add_warning_message_to_messenger(code=exc.code, message=exc.message)

        if SETTINGS.EXCEPTION_HANDLER.HANDLING:
            return self._handle_exception_gando_handling_true(exc)
        return self._handle_exception_gando_handling_false(exc)

    def exception_handler(self, exc, context):
        """
        Returns the response that should be used for any given exception.

        By default, we handle the REST framework `APIException`, and also
        Django's built-in `Http404` and `PermissionDenied` exceptions.

        Any unhandled exceptions may return `None`, which will cause a 500 error
        to be raised.
        """
        if isinstance(exc, Http404):
            exc = exceptions.NotFound(*(exc.args))
        elif isinstance(exc, PermissionDenied):
            exc = exceptions.PermissionDenied(*(exc.args))

        if isinstance(exc, exceptions.APIException):
            headers = {}
            if getattr(exc, 'auth_header', None):
                headers['WWW-Authenticate'] = exc.auth_header
            if getattr(exc, 'wait', None):
                headers['Retry-After'] = '%d' % exc.wait

            self._exception_handler_messages(exc.detail)

            set_rollback()
            return Response(status=exc.status_code, headers=headers)

        return None

    def _exception_handler_messages(self, msg, base_key=None):
        if isinstance(msg, list):
            for e in msg:
                self._exception_handler_messages(e)
        elif isinstance(msg, dict):
            for k, v in msg.items():
                self._exception_handler_messages(v, base_key=k)
        else:
            key = msg.code if hasattr(msg, 'code') else 'e'
            key = base_key + '__' + key if base_key else key
            self.set_error_message(key=key, value=str(msg))

    def _handle_exception_gando_handling_true(self, exc):
        """
        Handle any exception that occurs, by returning an appropriate response,
        or re-raising the error.
        """
        if isinstance(exc, (exceptions.NotAuthenticated,
                            exceptions.AuthenticationFailed)):
            auth_header = self.get_authenticate_header(self.request)

            if auth_header:
                exc.auth_header = auth_header
            else:
                exc.status_code = status.HTTP_403_FORBIDDEN

        context = self.get_exception_handler_context()
        response = self.exception_handler(exc, context)

        if response is None:
            response = Response()

        self.set_exception_message(
            key='unexpectedError',
            value=exc.args
        )
        self.set_error_message(
            key='unexpectedError',
            value=(
                "An unexpected error has occurred based on your request type.\n"
                "Please do not repeat this request without changing your request.\n"
                "Be sure to read the documents on how to use this service correctly.\n"
                "In any case, discuss the issue with software support.\n"
            )
        )
        self.set_warning_message(
            key='unexpectedError',
            value='Please discuss this matter with software support.',
        )
        if SETTINGS.EXCEPTION_HANDLER.COMMUNICATION_WITH_SOFTWARE_SUPPORT:
            self.set_info_message(
                key='unexpectedError',
                value=(
                    f"Please share this problem with our technical experts"
                    f" at the Email address "
                    f"'{SETTINGS.EXCEPTION_HANDLER.COMMUNICATION_WITH_SOFTWARE_SUPPORT}'."
                ),
            )
        self.set_status_code(421)
        response.exception = True
        return response

    def _handle_exception_gando_handling_false(self, exc):
        """
        Handle any exception that occurs, by returning an appropriate response,
        or re-raising the error.
        """
        if isinstance(exc, (exceptions.NotAuthenticated,
                            exceptions.AuthenticationFailed)):
            # WWW-Authenticate header for 401 responses, else coerce to 403
            auth_header = self.get_authenticate_header(self.request)

            if auth_header:
                exc.auth_header = auth_header
            else:
                exc.status_code = status.HTTP_403_FORBIDDEN

        context = self.get_exception_handler_context()
        response = self.exception_handler(exc, context)

        if response is None:
            self.raise_uncaught_exception(exc)

        response.exception = True
        return response

    def finalize_response(self, request, response, *args, **kwargs):
        if isinstance(response, Response):
            self.helper()

            tmp = response.template_name if hasattr(response, 'template_name') else None
            template_name = tmp

            tmp = response.headers if hasattr(response, 'headers') else None
            headers = self.get_headers(tmp)

            tmp = response.exception if hasattr(response, 'exception') else None
            exception = self.get_exception_status(tmp)

            tmp = response.content_type if hasattr(response, 'content_type') else None
            content_type = tmp

            tmp = response.status_code if hasattr(response, 'status_code') else None
            status_code = self.get_status_code(tmp)

            tmp = response.data if hasattr(response, 'data') else None
            data = self.response_context(tmp)

            response = Response(
                data=data,
                status=status_code,
                template_name=template_name,
                headers=headers,
                exception=exception,
                content_type=content_type,
            )

            if self.__cookies_for_delete:
                for i in self.__cookies_for_delete:
                    response.delete_cookie(i)

            if self.__cookies_for_set:
                for i in self.__cookies_for_set:
                    response.set_cookie(**i)

        return super().finalize_response(request, response, *args, **kwargs)

    def response_context(self, data=None):
        self.response_schema_version = self.request.headers.get('Response-Schema-Version') or '1.0.0'
        if self.response_schema_version == '2.0.0':
            return self._response_context_v_2_0_0_response(data)
        return self._response_context_v_1_0_0_response(data)

    def _response_context_v_1_0_0_response(self, data=None):
        self.__data = self.__set_messages_from_data(data)

        status_code = self.get_status_code()
        content_type = self.get_content_type()
        data = self.validate_data()
        many = self.__many()

        monitor = self.__monitor

        has_warning = self.__has_warning()
        exception_status = self.get_exception_status()

        messages = self.__messages()

        success = self.__success()

        headers = self.get_headers()

        tmp = {
            'success': success,

            'status_code': status_code,

            'has_warning': has_warning,

            'monitor': self.monitor_play(monitor),

            'messenger': self.__messenger,

            'many': many,
            'data': data,
        }
        if self.__development_messages_display():
            tmp['development_messages'] = messages

        if self.__exception_status_display():
            tmp['exception_status'] = exception_status

        ret = tmp
        return ret

    def _response_context_v_2_0_0_response(self, data=None):
        self.__data = self.__set_messages_from_data(data)

        status_code = self.get_status_code()
        content_type = self.get_content_type()
        data = self.validate_data_v_2_0_0_response()
        many = self.__many_v_2_0_0_response()

        monitor = self.__monitor

        has_warning = self.__has_warning()
        exception_status = self.get_exception_status()

        messages = self.__messages()

        success = self.__success()

        headers = self.get_headers()

        tmp = {
            'success': success,
            'status_code': status_code,
            'messenger': self.__messenger,
        }
        tmp.update(data)

        if self.__development_messages_display():
            tmp['development_messages'] = messages

        if self.__exception_status_display():
            tmp['exception_status'] = exception_status

        ret = tmp
        return ret

    def __add_to_messenger(self, message, code, type_):
        self.__messenger.append(
            {
                'type': type_,
                'code': code,
                'message': message,
            }
        )

    def add_fail_message_to_messenger(self, message, code):
        self.__add_to_messenger(
            message=message,
            code=code,
            type_='FAIL',
        )

    def add_error_message_to_messenger(self, message, code):
        self.__add_to_messenger(
            message=message,
            code=code,
            type_='ERROR',
        )

    def add_warning_message_to_messenger(self, message, code):
        self.__add_to_messenger(
            message=message,
            code=code,
            type_='WARNING',
        )

    def add_success_message_to_messenger(self, message, code):
        self.__add_to_messenger(
            message=message,
            code=code,
            type_='SUCCESS',
        )

    def set_log_message(self, key, value):
        log = {key: value}
        self.__logs_message.append(log)

    def set_info_message(self, key, value):
        info = {key: value}
        self.__infos_message.append(info)

    def set_warning_message(self, key, value):
        warning = {key: value}
        self.__warnings_message.append(warning)

    def set_error_message(self, key, value):
        error = {key: value}
        self.__errors_message.append(error)

    def set_exception_message(self, key, value):
        exception = {key: value}
        self.__exceptions_message.append(exception)

    def set_headers(self, key, value):
        self.__headers[key] = value

    def get_headers(self, value: dict = None):
        if value:
            for k, v in value.items():
                self.set_headers(k, v)
        return self.__headers

    def __messages(self, ) -> dict:
        tmp = {
            'info': self.__infos_message,
            'warning': self.__warnings_message,
            'error': self.__errors_message,

        }
        if self.__debug_status:
            tmp['log'] = self.__logs_message
            tmp['exception'] = self.__exceptions_message

        ret = tmp
        return ret

    def __many(self):
        if isinstance(self.__data, list):
            return True
        if (
            isinstance(self.__data, dict) and
            'count' in self.__data and
            'next' in self.__data and
            'previous' in self.__data and
            'results' in self.__data
        ):
            return True
        return False

    def __many_v_2_0_0_response(self):
        if isinstance(self.__data, list):
            return True
        if (
            isinstance(self.__data, dict) and
            'count' in self.__data and
            'next' in self.__data and
            'previous' in self.__data and
            'result' in self.__data
        ):
            return True
        return False

    def __fail_message_messenger(self):
        for msg in self.__messenger:
            if msg.get('type', '') == 'FAIL' or msg.get('type', '') == 'ERROR':
                return True
        return False

    def __warning_message_messenger(self):
        for msg in self.__messenger:
            if msg.get('type', '') == 'WARNING':
                return True
        return False

    def __success(self):
        if (
            len(self.__errors_message) == 0 and
            len(self.__exceptions_message) == 0 and
            not self.__exception_status and
            not self.__fail_message_messenger()
        ):
            return True
        return False

    def __has_warning(self):
        if len(self.__warnings_message) != 0 and self.__warning_message_messenger():
            return True
        return False

    def set_status_code(self, value: int):
        self.__status_code = value

    def get_status_code(self, value: int = None):
        if value and 100 <= value < 600 and value != 200:
            self.set_status_code(value)

        return self.__status_code or 200

    def set_content_type(self, value: str):
        self.__content_type = value

    def get_content_type(self, value: str = None):
        if value:
            self.set_content_type(value)

        return self.__content_type

    def set_exception_status(self, value: bool):
        self.__exception_status = value

    def get_exception_status(self, value: bool = None):
        if value is not None:
            self.set_exception_status(value)

        return self.__exception_status

    def set_monitor(self, key, value):
        if key in self.__allowed_monitor_keys:
            self.__monitor[key] = value

    def __monitor_func_loader(self, f, *args, **kwargs):
        try:
            mod_name, func_name = f.rsplit('.', 1)
            mod = importlib.import_module(mod_name)
            func = getattr(mod, func_name)

            return func(request=self.request, *args, **kwargs)
        except PassException as exc:
            frame_info = getframeinfo(currentframe())
            self.set_log_message(
                key='pass',
                value=f"message:{exc.args[0]}, "
                      f"file_name: {frame_info.filename}, "
                      f"line_number: {frame_info.lineno}")
            return None

    def monitor_play(self, monitor=None, *args, **kwargs):
        monitor = monitor or {}
        for key, f in SETTINGS.MONITOR.items():
            monitor[key] = self.__monitor_func_loader(f, *args, **kwargs)

        return monitor

    @property
    def __allowed_monitor_keys(self):
        return SETTINGS.MONITOR_KEYS

    def validate_data(self):
        data = self.__data

        if data is None:
            self.__set_default_message()
            tmp = {'result': {}}

        elif isinstance(data, str) or issubclass(type(data), str):
            data = self.__set_dynamic_message(data)
            tmp = {'result': {'string': data} if data else {}}

        elif isinstance(data, list):
            tmp = {
                'count': len(data),
                'next': None,
                'previous': None,
                'results': data,
            }

        elif isinstance(data, dict):
            tmp = {'result': data}

        else:
            tmp = {'result': {}}

        ret = tmp
        return ret

    def validate_data_v_2_0_0_response(self):
        data = self.__data

        if data is None:
            self.__set_default_message()
            tmp = {'result': {}}

        elif isinstance(data, str) or issubclass(type(data), str):
            data = self.__set_dynamic_message(data)
            tmp = {'result': {'string': data} if data else {}}

        elif isinstance(data, list):
            tmp = {
                'count': len(data),
                'next': None,
                'previous': None,
                'result': data,
            }

        elif isinstance(data, dict):
            tmp = {'result': data}

        else:
            tmp = {'result': {}}

        ret = tmp
        return ret

    @property
    def __debug_status(self):
        return SETTINGS.DEBUG

    @property
    def __development_state(self):
        return SETTINGS.DEVELOPMENT_STATE

    def __development_messages_display(self):
        if self.__development_state:
            ret = self.request.headers.get('Development-Messages-Display', 'True') == 'True'
        else:
            ret = False
        return ret

    def __exception_status_display(self):
        if self.__development_state:
            ret = self.request.headers.get('Exception-Status-Display', 'True') == 'True'
        else:
            ret = False
        return ret

    def response(self, output_data=None):
        data = output_data
        rsp = Response(
            data,
            status=self.get_status_code(),
            headers=self.get_headers(),
        )

        return rsp

    def get_host(self):
        return self.request._request._current_scheme_host

    def append_host_to_url(self, value):
        ret = f'{self.get_host()}{value}'
        return ret

    @staticmethod
    def get_media_url():
        from django.conf import settings

        ret = settings.MEDIA_URL
        return ret

    def convert_filename_to_url(self, file_name):
        if file_name is None:
            return None
        ret = f'{self.get_media_url()}{file_name}'
        return ret

    def convert_filename_to_url_localhost(self, file_name):
        if file_name is None:
            return None
        ret = f'{self.get_host()}{self.get_media_url()}{file_name}'
        return ret

    def helper(self):
        pass

    def __default_message(self):
        status_code = self.get_status_code()

        if 100 <= status_code < 200:
            msg = 'please wait...'

        elif 200 <= status_code < 300:
            if status_code == 201:
                msg = 'The desired object was created correctly.'
            else:
                msg = 'Your request has been successfully registered.'

        elif 300 <= status_code < 400:
            msg = 'The requirements for your request are not available.'

        elif 400 <= status_code < 500:
            if status_code == 400:
                msg = 'Bad Request...'

            elif status_code == 401:
                msg = 'Your authentication information is not available.'

            elif status_code == 403:
                msg = 'You do not have access to this section.'

            elif status_code == 404:
                msg = 'There is no information about your request.'

            elif status_code == 421:
                msg = (
                    "An unexpected error has occurred based on your request type.\n"
                    "Please do not repeat this request without changing your request.\n"
                    "Be sure to read the documents on how to use this service correctly.\n"
                    "In any case, discuss the issue with software support.\n"
                )

            else:
                msg = 'There was an error in how to send the request.'

        elif 500 <= status_code:
            msg = 'The server is unable to respond to your request.'

        else:
            msg = 'Undefined.'

        return msg

    def __default_messenger_message_adder(self):
        status_code = self.get_status_code()

        if 100 <= status_code < 200:
            self.__add_to_messenger(
                message=DefaultResponse100FailMessage.message,
                code=DefaultResponse100FailMessage.code,
                type_=DefaultResponse100FailMessage.type,
            )

        elif 200 <= status_code < 300:
            if status_code == 201:
                self.__add_to_messenger(
                    message=DefaultResponse201SuccessMessage.message,
                    code=DefaultResponse201SuccessMessage.code,
                    type_=DefaultResponse201SuccessMessage.type,
                )
            else:
                self.__add_to_messenger(
                    message=DefaultResponse200SuccessMessage.message,
                    code=DefaultResponse200SuccessMessage.code,
                    type_=DefaultResponse200SuccessMessage.type,
                )

        elif 300 <= status_code < 400:
            self.__add_to_messenger(
                message=DefaultResponse300FailMessage.message,
                code=DefaultResponse300FailMessage.code,
                type_=DefaultResponse300FailMessage.type,
            )

        elif 400 <= status_code < 500:
            if status_code == 400:
                self.__add_to_messenger(
                    message=DefaultResponse400FailMessage.message,
                    code=DefaultResponse400FailMessage.code,
                    type_=DefaultResponse400FailMessage.type,
                )

            elif status_code == 401:
                self.__add_to_messenger(
                    message=DefaultResponse401FailMessage.message,
                    code=DefaultResponse401FailMessage.code,
                    type_=DefaultResponse401FailMessage.type,
                )

            elif status_code == 403:
                self.__add_to_messenger(
                    message=DefaultResponse403FailMessage.message,
                    code=DefaultResponse403FailMessage.code,
                    type_=DefaultResponse403FailMessage.type,
                )

            elif status_code == 404:
                self.__add_to_messenger(
                    message=DefaultResponse404FailMessage.message,
                    code=DefaultResponse404FailMessage.code,
                    type_=DefaultResponse404FailMessage.type,
                )

            elif status_code == 421:
                self.__add_to_messenger(
                    message=DefaultResponse421FailMessage.message,
                    code=DefaultResponse421FailMessage.code,
                    type_=DefaultResponse421FailMessage.type,
                )

            else:
                self.__add_to_messenger(
                    message=DefaultResponse400FailMessage.message,
                    code=DefaultResponse400FailMessage.code,
                    type_=DefaultResponse400FailMessage.type,
                )

        elif 500 <= status_code:
            self.__add_to_messenger(
                message=DefaultResponse500FailMessage.message,
                code=DefaultResponse500FailMessage.code,
                type_=DefaultResponse500FailMessage.type,
            )

        else:
            pass

    def __set_default_message(self):
        self.__default_messenger_message_adder()

        status_code = self.get_status_code()

        if 100 <= status_code < 200:
            self.set_warning_message('status_code_1xx', self.__default_message())

        elif 200 <= status_code < 300:
            self.set_info_message('status_code_2xx', self.__default_message())

        elif 300 <= status_code < 400:
            self.set_error_message('status_code_3xx', self.__default_message())

        elif 400 <= status_code < 500:
            self.set_error_message('status_code_4xx', self.__default_message())

        elif 500 <= status_code:
            self.set_error_message('status_code_5xx', self.__default_message())

        else:
            self.set_error_message('status_code_xxx', self.__default_message())

    def __set_messages_from_data(self, data):
        if isinstance(data, str) or issubclass(type(data), str):
            return self.__set_dynamic_message(data)

        if isinstance(data, list):
            tmp = []
            for i in data:
                rslt = self.__set_messages_from_data(i)
                # if rslt:
                tmp.append(rslt)

            ret = tmp
            return ret

        if isinstance(data, dict):
            tmp = {}
            for k, v in data.items():
                rslt = self.__set_messages_from_data(v)
                # if rslt is not None:
                tmp[k] = v

            ret = tmp
            return ret

        return data

    def __set_dynamic_message(self, value):
        if isinstance(value, InfoStringMessage):
            self.set_info_message(key=value.code, value=value)
            return None
        if isinstance(value, ErrorStringMessage) or isinstance(value, ErrorDetail):
            self.set_error_message(key=value.code, value=value)
            return None
        if isinstance(value, WarningStringMessage):
            self.set_warning_message(key=value.code, value=value)
            return None
        if isinstance(value, LogStringMessage):
            self.set_log_message(key=value.code, value=value)
            return None
        if isinstance(value, ExceptionStringMessage):
            self.set_exception_message(key=value.code, value=value)
            return None

        return value

    class Cookie(BaseModel):
        key: str
        value: Any = ""
        max_age: Any = None
        expires: Any = None
        path: Any = "/"
        domain: Any = None
        secure: Any = False
        httponly: Any = False
        samesite: Any = None

    def cookie_getter(self, key: str):
        ret = self.request.COOKIES.get(key)
        return ret

    def cookie_setter(self, key: str, **kwargs):
        ret = self.__cookies_for_set.append(self.Cookie(key=key, **kwargs).model_dump())
        return ret

    def cookie_deleter(self, key: str):
        ret = self.__cookies_for_delete.append(key)
        return ret


class CreateAPIView(BaseAPI, DRFGCreateAPIView):
    def create(self, request, *args, **kwargs):
        if hasattr(self, 'check_validate_user') and self.check_validate_user:
            user_lookup_field = 'id'
            if hasattr(self, 'user_lookup_field'):
                user_lookup_field = self.user_lookup_field
            if not _valid_user(request=request, user_id=kwargs.get(user_lookup_field)):
                return Response(status=403)

        data = request.data.copy()
        user_field_name = 'user'
        if hasattr(self, 'user_field_name'):
            user_field_name = self.user_field_name
        data[user_field_name] = request.user.id

        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class ListAPIView(BaseAPI, DRFGListAPIView):
    def get_queryset(self):
        qs = super().get_queryset()
        if hasattr(self, 'for_user') and self.for_user:
            qs = qs.filter(user_id=self.request.user.id)
        return qs

    def get(self, request, *args, **kwargs):
        if hasattr(self, 'check_validate_user') and self.check_validate_user:
            user_lookup_field = 'id'
            if hasattr(self, 'user_lookup_field'):
                user_lookup_field = self.user_lookup_field
            if not _valid_user(request=request, user_id=kwargs.get(user_lookup_field)):
                return Response(status=403)
        return super().get(request, *args, **kwargs)


class RetrieveAPIView(BaseAPI, DRFGRetrieveAPIView):
    def get_queryset(self):
        qs = super().get_queryset()
        if hasattr(self, 'for_user') and self.for_user:
            qs = qs.filter(user_id=self.request.user.id)
        return qs

    def get(self, request, *args, **kwargs):
        if hasattr(self, 'check_validate_user') and self.check_validate_user:
            user_lookup_field = 'id'
            if hasattr(self, 'user_lookup_field'):
                user_lookup_field = self.user_lookup_field
            if not _valid_user(request=request, user_id=kwargs.get(user_lookup_field)):
                return Response(status=403)
        return super().get(request, *args, **kwargs)


class UpdateAPIView(BaseAPI, DRFGUpdateAPIView):
    def get_queryset(self):
        qs = super().get_queryset()
        if hasattr(self, 'for_user') and self.for_user:
            qs = qs.filter(user_id=self.request.user.id)
        return qs

    def update(self, request, *args, **kwargs):
        if hasattr(self, 'check_validate_user') and self.check_validate_user:
            user_lookup_field = 'id'
            if hasattr(self, 'user_lookup_field'):
                user_lookup_field = self.user_lookup_field
            if not _valid_user(request=request, user_id=kwargs.get(user_lookup_field)):
                return Response(status=403)

        partial = kwargs.pop('partial', False)
        instance = self.get_object()

        data = request.data.copy()
        user_field_name = 'user'
        if hasattr(self, 'user_field_name'):
            user_field_name = self.user_field_name
        data[user_field_name] = request.user.id

        serializer = self.get_serializer(instance, data=data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return Response(serializer.data)


class DestroyAPIView(BaseAPI, DRFGDestroyAPIView):
    def get_queryset(self):
        qs = super().get_queryset()
        if hasattr(self, 'for_user') and self.for_user:
            qs = qs.filter(user_id=self.request.user.id)
        return qs

    def delete(self, request, *args, **kwargs):
        if hasattr(self, 'check_validate_user') and self.check_validate_user:
            user_lookup_field = 'id'
            if hasattr(self, 'user_lookup_field'):
                user_lookup_field = self.user_lookup_field
            if not _valid_user(request=request, user_id=kwargs.get(user_lookup_field)):
                return Response(status=403)
        return self.destroy(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_destroy(instance=instance, soft_delete=kwargs.get('soft_delete', False))
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, instance, soft_delete=False):
        if instance is None:
            return

        if soft_delete:
            instance.available = 0
            instance.save()
        else:
            instance.delete()



from django.contrib import admin
from django.utils.translation import gettext_lazy as _


def verbose_name(value: str):
    tmp = value[0].upper()
    i = 1
    while i < len(value):
        if value[i] != '_':
            tmp += value[i]
        else:
            tmp += ' '
            i += 1
            tmp += value[i].upper()
        i += 1
    ret = tmp
    return ret


class BaseModelAdmin(admin.ModelAdmin):

    def __init__(self, *args, **kwargs):
        # Display Management
        self.available_field_in_display_list = True
        self.id_field_in_display_list = True
        self.created_dt_field_in_display_list = True
        self.updated_dt_field_in_display_list = True

        self.list_display = [] if not self.list_display else self.list_display
        self.list_display_links = [] if not self.list_display_links else self.list_display_links

        # Filter Fields Management
        self.available_field_in_filter_list = True
        self.created_dt_field_in_filter_list = True
        self.updated_dt_field_in_filter_list = True
        self.list_filter = [] if not self.list_filter else self.list_filter

        # Search Fields Management
        self.id_field_in_search_fields = True
        self.search_fields = [] if not self.search_fields else self.search_fields

        # Readonly Fields Management
        self.id_field_is_readonly = True
        self.available_field_is_readonly = True
        self.created_dt_field_is_readonly = True
        self.updated_dt_field_is_readonly = True
        self.readonly_fields = [] if not self.readonly_fields else self.readonly_fields

        super().__init__(*args, **kwargs)

    _list_display = []

    @property
    def list_display(self):
        return self._list_display

    @list_display.setter
    def list_display(self, value):
        value = list(value)

        available_field_in_display_list = (
            self.available_field_in_display_list
            if hasattr(self, 'available_field_in_display_list') else False)
        id_field_in_display_list = (
            self.id_field_in_display_list
            if hasattr(self, 'id_field_in_display_list') else False)
        created_dt_field_in_display_list = (
            self.created_dt_field_in_display_list
            if hasattr(self, 'created_dt_field_in_display_list') else False)
        updated_dt_field_in_display_list = (
            self.updated_dt_field_in_display_list
            if hasattr(self, 'updated_dt_field_in_display_list') else False)

        tmp = ['available'] if 'available' not in value and available_field_in_display_list else []
        tmp += ['id'] if 'id' not in value and id_field_in_display_list else []
        tmp += value
        tmp += ['created_dt'] if 'created_dt' not in value and created_dt_field_in_display_list else []
        tmp += ['updated_dt'] if 'updated_dt' not in value and updated_dt_field_in_display_list else []

        self._list_display = tmp

    _list_display_links = []

    @property
    def list_display_links(self):
        return self._list_display_links

    @list_display_links.setter
    def list_display_links(self, value):
        value = list(value)

        available_field_in_display_list = (
            self.available_field_in_display_list
            if hasattr(self, 'available_field_in_display_list') else False)
        id_field_in_display_list = (
            self.id_field_in_display_list
            if hasattr(self, 'id_field_in_display_list') else False)
        created_dt_field_in_display_list = (
            self.created_dt_field_in_display_list
            if hasattr(self, 'created_dt_field_in_display_list') else False)
        updated_dt_field_in_display_list = (
            self.updated_dt_field_in_display_list
            if hasattr(self, 'updated_dt_field_in_display_list') else False)

        tmp = ['available'] if 'available' not in value and available_field_in_display_list else []
        tmp += ['id'] if 'id' not in value and id_field_in_display_list else []
        tmp += value
        tmp += ['created_dt'] if 'created_dt' not in value and created_dt_field_in_display_list else []
        tmp += ['updated_dt'] if 'updated_dt' not in value and updated_dt_field_in_display_list else []

        self._list_display_links = tmp

    image_fields_name_list = []

    def __set_image_fieldsets(self):
        tmp = []
        for i in self.image_fields_name_list:
            tmp += [
                (f'{i}_category', f'{i}_device_type',),
                (f'{i}_alt', f'{i}_src',),
                (f'{i}_width', f'{i}_height',),
                f'{i}_description',
                f'{i}_blurbase64',
            ]

        ret = [('Images', {'fields': tmp})] if tmp else []
        return ret

    def __get_image_read_only_fields(self):
        ret = []
        for i in self.image_fields_name_list:
            ret.append(f'{i}_blurbase64')
            ret.append(f'{i}_width')
            ret.append(f'{i}_height')
        return ret

    _fieldsets = []

    @property
    def fieldsets(self):
        return self._fieldsets

    @fieldsets.setter
    def fieldsets(self, value):
        tmp = [(_('Initial'), {'fields': [('available', 'id',)]})]
        tmp += value
        tmp += self.__set_image_fieldsets()
        tmp += [(_('Extra'), {'fields': [('created_dt', 'updated_dt',)]})]

        self._fieldsets = tmp

    _readonly_fields = []

    @property
    def readonly_fields(self):
        return self._readonly_fields

    @readonly_fields.setter
    def readonly_fields(self, value):
        value = list(value)
        id_field_is_readonly = (
            self.id_field_is_readonly
            if hasattr(self, 'id_field_is_readonly') else False
        )
        available_field_is_readonly = (
            self.available_field_is_readonly
            if hasattr(self, 'available_field_is_readonly') else False
        )
        created_dt_field_is_readonly = (
            self.created_dt_field_is_readonly
            if hasattr(self, 'created_dt_field_is_readonly') else False
        )
        updated_dt_field_is_readonly = (
            self.updated_dt_field_is_readonly
            if hasattr(self, 'updated_dt_field_is_readonly') else False
        )

        tmp = ['id'] if 'id' not in value and id_field_is_readonly else []
        tmp += value
        tmp += self.__get_image_read_only_fields()
        tmp += ['available'] if 'available' not in value and available_field_is_readonly else []
        tmp += ['created_dt'] if 'created_dt' not in value and created_dt_field_is_readonly else []
        tmp += ['updated_dt'] if 'updated_dt' not in value and updated_dt_field_is_readonly else []

        self._readonly_fields = tmp

    _list_filter = []

    @property
    def list_filter(self):
        return self._list_filter

    @list_filter.setter
    def list_filter(self, value):
        value = list(value)

        available_field_in_filter_list = (
            self.available_field_in_filter_list
            if hasattr(self, 'available_field_in_filter_list') else False)
        created_dt_field_in_filter_list = (
            self.created_dt_field_in_filter_list
            if hasattr(self, 'created_dt_field_in_filter_list') else False)
        updated_dt_field_in_filter_list = (
            self.updated_dt_field_in_filter_list
            if hasattr(self, 'updated_dt_field_in_filter_list') else False)

        tmp = ['available'] if 'available' not in value and available_field_in_filter_list else []
        tmp += value
        tmp += ['created_dt'] if 'created_dt' not in value and created_dt_field_in_filter_list else []
        tmp += ['updated_dt'] if 'updated_dt' not in value and updated_dt_field_in_filter_list else []

        self._list_filter = tmp

    _search_fields = []

    @property
    def search_fields(self):
        return self._search_fields

    @search_fields.setter
    def search_fields(self, value):
        value = list(value)

        id_field_in_search_fields = (
            self.id_field_in_search_fields if hasattr(self, 'id_field_in_search_fields') else False)

        tmp = ['id'] if 'id' not in value and id_field_in_search_fields else []
        tmp += value

        self._search_fields = tmp



from django.utils.translation import gettext_lazy as _
from django.contrib.auth import get_user_model
from django.db import models


def user_agent_device_key_creator(instance):
    from hashlib import md5

    str_ = (
        f"{instance.user_agent_is_mobile}-"
        f"{instance.user_agent_is_tablet}-"
        f"{instance.user_agent_is_touch_capable}-"
        f"{instance.user_agent_is_pc}-"
        f"{instance.user_agent_is_bot}-"
        f"{instance.user_agent_browser_family}-"
        f"{instance.user_agent_browser_version}-"
        f"{instance.user_agent_os_family}-"
        f"{instance.user_agent_os_version}-"
        f"{instance.user_agent_device_family}-"
        f"{instance.user_agent_device_brand}-"
        f"{instance.user_agent_device_model}-"
        f"{instance.ip}"
    )
    ret = str(md5(str_.encode('utf-8')).hexdigest())
    return ret


class UserAgentDevice(models.Model):
    key = models.CharField(
        verbose_name=_('Key'),
        max_length=255,
        blank=True,
        null=True,
        unique=True,
        db_index=True,
    )
    user = models.ForeignKey(
        verbose_name=_('User'),
        to=get_user_model(),
        on_delete=models.CASCADE,
        related_name='uad',
        db_index=True,
        blank=True,
        null=True,
    )
    user_agent_is_mobile = models.IntegerField(
        verbose_name=_('Mobile'),
        blank=True,
        null=True,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
    )
    user_agent_is_tablet = models.IntegerField(
        verbose_name=_('Tablet'),
        blank=True,
        null=True,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
    )
    user_agent_is_touch_capable = models.IntegerField(
        verbose_name=_('TouchCapable'),
        blank=True,
        null=True,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
    )
    user_agent_is_pc = models.IntegerField(
        verbose_name=_('PC'),
        blank=True,
        null=True,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
    )
    user_agent_is_bot = models.IntegerField(
        verbose_name=_('Bot'),
        blank=True,
        null=True,
        choices=(
            (0, 'No'),
            (1, 'Yes'),
        ),
    )
    user_agent_browser_family = models.CharField(
        verbose_name=_('BrowserFamily'),
        blank=True,
        null=True,
        max_length=255,
    )
    user_agent_browser_version = models.CharField(
        verbose_name=_('BrowserVersion'),
        blank=True,
        null=True,
        max_length=255,
    )
    user_agent_os_family = models.CharField(
        verbose_name=_('OSFamily'),
        blank=True,
        null=True,
        max_length=255,
    )
    user_agent_os_version = models.CharField(
        verbose_name=_('OSVersion'),
        blank=True,
        null=True,
        max_length=255,
    )
    user_agent_device_family = models.CharField(
        verbose_name=_('DeviceFamily'),
        blank=True,
        null=True,
        max_length=255,
    )
    user_agent_device_brand = models.CharField(
        verbose_name=_('DeviceBrand'),
        blank=True,
        null=True,
        max_length=255,
    )
    user_agent_device_model = models.CharField(
        verbose_name=_('DeviceModel'),
        blank=True,
        null=True,
        max_length=255,
    )
    ip = models.CharField(
        verbose_name=_('IP'),
        max_length=255,
        blank=True,
        null=True,
    )
    created_dt = models.DateTimeField(
        verbose_name=_('Created Datetime'),
        auto_now_add=True,
        db_index=True,
    )

    def __str__(self):
        return f'{self.key}'

    def save(self, **kwargs):
        self.key = user_agent_device_key_creator(self)
        super().save(**kwargs)

    class Meta:
        verbose_name = _('UserAgentDevice')
        verbose_name_plural = _('UserAgentDevices')



from rest_framework.exceptions import APIException

DEVELOPER_EXCEPTION = 'developer_exception'
DEVELOPER_ERROR = 'developer_error'
DEVELOPER_WARNING = 'developer_warning'
ENDUSER_FAIL = 'enduser_fail'
ENDUSER_ERROR = 'enduser_error'
ENDUSER_WARNING = 'enduser_warning'


class ResponseAPIMessage(APIException):
    def __init__(self, message=None, code=None, typ=None, status_code=500):
        super().__init__(detail=message, code=code)
        self.typ = typ
        self.status_code = status_code
        self.code = code
        self.message = message


class DeveloperResponseAPIMessage(ResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class DeveloperExceptionResponseAPIMessage(DeveloperResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class DeveloperErrorResponseAPIMessage(DeveloperResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class DeveloperWarningResponseAPIMessage(DeveloperResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class DeveloperExceptionResponseAPIMessageBadRequest(DeveloperExceptionResponseAPIMessage):
    def __init__(self, message='Bad Request', code='bad_request'):
        super().__init__(
            message=message,
            code=code,
            typ=DEVELOPER_EXCEPTION,
            status_code=400,
        )


class DeveloperErrorResponseAPIMessageBadRequest(DeveloperErrorResponseAPIMessage):
    def __init__(self, message='Bad Request', code='bad_request'):
        super().__init__(
            message=message,
            code=code,
            typ=DEVELOPER_ERROR,
            status_code=400,
        )


class DeveloperWarningResponseAPIMessageBadRequest(DeveloperWarningResponseAPIMessage):
    def __init__(self, message='Bad Request', code='bad_request'):
        super().__init__(
            message=message,
            code=code,
            typ=DEVELOPER_WARNING,
            status_code=400,
        )


class DeveloperExceptionResponseAPIMessageForbidden403(DeveloperExceptionResponseAPIMessage):
    def __init__(
        self,
        message="You do not have permission to access this resource. "
                "Please do not try again, "
                "or if you doubt the authenticity of the requested permissions, "
                "try again after checking the authenticity of the permissions.",
        code='forbidden',
    ):
        super().__init__(
            message=message,
            code=code,
            typ=DEVELOPER_EXCEPTION,
            status_code=403,
        )


class DeveloperErrorResponseAPIMessageForbidden403(DeveloperErrorResponseAPIMessage):
    def __init__(
        self,
        message="You do not have permission to access this resource. "
                "Please do not try again, "
                "or if you doubt the authenticity of the requested permissions, "
                "try again after checking the authenticity of the permissions.",
        code='forbidden',
    ):
        super().__init__(
            message=message,
            code=code,
            typ=DEVELOPER_ERROR,
            status_code=403,
        )


class DeveloperWarningResponseAPIMessageForbidden403(DeveloperWarningResponseAPIMessage):
    def __init__(
        self,
        message="You do not have permission to access this resource. "
                "Please do not try again, "
                "or if you doubt the authenticity of the requested permissions, "
                "try again after checking the authenticity of the permissions.",
        code='forbidden',
    ):
        super().__init__(
            message=message,
            code=code,
            typ=DEVELOPER_WARNING,
            status_code=403,
        )


class EnduserResponseAPIMessage(ResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class EnduserFailResponseAPIMessage(EnduserResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class EnduserErrorResponseAPIMessage(EnduserResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class EnduserWarningResponseAPIMessage(EnduserResponseAPIMessage):
    def __init__(self, message=None, code=None, typ=None, status_code=None):
        super().__init__(
            message=message,
            code=code,
            typ=typ,
            status_code=status_code,
        )


class EnduserFailResponseAPIMessageBadRequest(EnduserFailResponseAPIMessage):
    def __init__(self, message='Bad Request', code='bad_request'):
        super().__init__(
            message=message,
            code=code,
            typ=ENDUSER_FAIL,
            status_code=400,
        )


class EnduserErrorResponseAPIMessageBadRequest(EnduserErrorResponseAPIMessage):
    def __init__(self, message='Bad Request', code='bad_request'):
        super().__init__(
            message=message,
            code=code,
            typ=ENDUSER_ERROR,
            status_code=400,
        )


class EnduserWarningResponseAPIMessageBadRequest(EnduserWarningResponseAPIMessage):
    def __init__(self, message='Bad Request', code='bad_request'):
        super().__init__(
            message=message,
            code=code,
            typ=ENDUSER_WARNING,
            status_code=400,
        )


class EnduserFailResponseAPIMessageForbidden403(EnduserFailResponseAPIMessage):
    def __init__(
        self,
        message='You do not have permission to access this resource. Please do not try again.',
        code='forbidden',
    ):
        super().__init__(
            message=message,
            code=code,
            typ=ENDUSER_FAIL,
            status_code=403,
        )


class EnduserErrorResponseAPIMessageForbidden403(EnduserErrorResponseAPIMessage):
    def __init__(
        self,
        message='You do not have permission to access this resource. Please do not try again.',
        code='forbidden',
    ):
        super().__init__(
            message=message,
            code=code,
            typ=ENDUSER_ERROR,
            status_code=403,
        )


class EnduserWarningResponseAPIMessageForbidden403(EnduserWarningResponseAPIMessage):
    def __init__(
        self,
        message='You do not have permission to access this resource. Please do not try again.',
        code='forbidden',
    ):
        super().__init__(
            message=message,
            code=code,
            typ=ENDUSER_WARNING,
            status_code=403,
        )



# Generated by Django 5.0.1 on 2024-02-04 00:33

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserAgentDevice',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(blank=True, db_index=True, max_length=255, null=True, unique=True, verbose_name='Key')),
                ('user_agent_is_mobile', models.IntegerField(blank=True, choices=[(0, 'No'), (1, 'Yes')], null=True, verbose_name='Mobile')),
                ('user_agent_is_tablet', models.IntegerField(blank=True, choices=[(0, 'No'), (1, 'Yes')], null=True, verbose_name='Tablet')),
                ('user_agent_is_touch_capable', models.IntegerField(blank=True, choices=[(0, 'No'), (1, 'Yes')], null=True, verbose_name='TouchCapable')),
                ('user_agent_is_pc', models.IntegerField(blank=True, choices=[(0, 'No'), (1, 'Yes')], null=True, verbose_name='PC')),
                ('user_agent_is_bot', models.IntegerField(blank=True, choices=[(0, 'No'), (1, 'Yes')], null=True, verbose_name='Bot')),
                ('user_agent_browser_family', models.CharField(blank=True, max_length=255, null=True, verbose_name='BrowserFamily')),
                ('user_agent_browser_version', models.CharField(blank=True, max_length=255, null=True, verbose_name='BrowserVersion')),
                ('user_agent_os_family', models.CharField(blank=True, max_length=255, null=True, verbose_name='OSFamily')),
                ('user_agent_os_version', models.CharField(blank=True, max_length=255, null=True, verbose_name='OSVersion')),
                ('user_agent_device_family', models.CharField(blank=True, max_length=255, null=True, verbose_name='DeviceFamily')),
                ('user_agent_device_brand', models.CharField(blank=True, max_length=255, null=True, verbose_name='DeviceBrand')),
                ('user_agent_device_model', models.CharField(blank=True, max_length=255, null=True, verbose_name='DeviceModel')),
                ('ip', models.CharField(blank=True, max_length=255, null=True, verbose_name='IP')),
                ('created_dt', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Created Datetime')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='uad', to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'UserAgentDevice',
                'verbose_name_plural': 'UserAgentDevices',
            },
        ),
    ]



from .__base import BaseService
from .__database_manager import BaseDataBaseManagerService

from .__creator import BaseCreatorService
from .__getter_creator import BaseGetterCreatorService

from .__updater import BaseUpdaterService

from .__getter import BaseGetterService



FLOAT_TYPE = 'Float'
INTEGER_TYPE = 'Integer'
STRING_TYPE = 'String'
UUID_TYPE = 'UUID'
BOOLEAN_TYPE = 'Boolean'
DATETIME_TYPE = 'DateTime'
DATE_TYPE = 'Datetime'
TIME_TYPE = 'Time'
EMAIL_TYPE = 'Email'
LINK_TYPE = 'Link'
PASSWORD_TYPE = 'Password'
USERNAME_TYPE = 'Username'
PROPORTION_TYPE = 'Proportion'
IMAGE_TYPE_Default = 'Image'
BLURBASE64_TYPE = 'BlurBase64'
TEXT_TYPE = 'Text'
IMAGE_TYPE = {
    'directory_name': STRING_TYPE,
    'alt': STRING_TYPE,
    'height': INTEGER_TYPE,
    'width': INTEGER_TYPE,
    'size': INTEGER_TYPE,
    'proportion': PROPORTION_TYPE,
    'file_format': STRING_TYPE,
    'src': IMAGE_TYPE_Default,
    'blurbase64': BLURBASE64_TYPE,
    'description': TEXT_TYPE
}


def response_template_render(**kwargs):
    return {
        'success': kwargs.get('success', True),
        'status_code': kwargs.get('status_code', 200),
        'has_warning': kwargs.get('has_warning', False),
        'exception_status': kwargs.get('exception_status', False),
        'monitor': kwargs.get('monitor', {}),
        'messenger': kwargs.get('messenger', []),
        'many': kwargs.get('many', False),
        'data': kwargs.get('data', {'result': {}}),
        'messages': {
            'warning': kwargs.get('warning', []),
            'error': kwargs.get('error', []),
        }
    }


def response_template_render_many_false(**kwargs):
    return response_template_render(**kwargs)


def response_template_render_many_true(**kwargs):
    kwargs['many'] = True
    kwargs['data'] = kwargs.get('data', {})
    kwargs['data']['count'] = INTEGER_TYPE
    kwargs['data']['next'] = LINK_TYPE
    kwargs['data']['previous'] = LINK_TYPE
    return response_template_render(**kwargs)


RESPONSE_BODY_401 = response_template_render(
    success=False, status_code=401, exception_status=True, error=[
        {
            'not_authenticated': 'Authentication credentials were not provided.'
        }
    ])



from gando.http.responses.schemas import ResponseSchema
from gando.http.requests.schemas import RequestSchema
from gando.http.requests.methods import (
    GET,
    POST,
    PUT,
    PATCH,
    DELETE,
    HEAD,
    OPTIONS,
    TRACE,
)


class Base:
    http_method_names = [
        GET,
        POST,
        PUT,
        PATCH,
        DELETE,
        HEAD,
        OPTIONS,
        TRACE,
    ]

    def __init__(self, request: RequestSchema, **kwargs):
        self.request: RequestSchema = request
        for key, value in kwargs.items():
            setattr(self, key, value)

    def execute(self, *args, **kwargs):
        return self.dispatch(self.request, *args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        if request.method.lower() in self.http_method_names:
            handler = getattr(
                self, request.method.lower(), self.http_method_not_allowed
            )
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)

    def http_method_not_allowed(self, request, *args, **kwargs):
        response = ResponseSchema(
            success=False,
            status_code=405,
            messages={
                'info': [{'allowed_methods': self._allowed_methods()}],
                'error': [{'method': 'Method Not Allowed.'}],
            },
        )

        return response

    def _allowed_methods(self):
        return [m.upper() for m in self.http_method_names if hasattr(self, m)]


class BaseInterface(Base):
    pass



import re

from django.utils.translation import gettext_lazy as _
from django.utils.deconstruct import deconstructible
from django.core import validators
from django.db import models

from gando.utils.converters.images import small_blur_base64
from gando.utils.uploaders.images import ImageUploadTo


def verbose_name(value: str):
    tmp = value[0].upper()
    i = 1
    while i < len(value):
        if value[i] != '_':
            tmp += value[i]
        else:
            tmp += ' '
            i += 1
            tmp += value[i].upper()
        i += 1
    ret = tmp
    return ret


class BlurBase64Field(models.TextField):
    def formfield(self, **kwargs):
        return None

    def pre_save(self, model_instance, add):
        _src = None
        if hasattr(model_instance, f'{self.PARENT_FIELD_NAME}_src'):
            _src = getattr(model_instance, f'{self.PARENT_FIELD_NAME}_src')
        if _src:
            setattr(model_instance, self.attname, small_blur_base64(_src.file.name))

        return super().pre_save(model_instance, add)


class BaseMultiplyField(models.Field):
    def __init__(self, **kwargs):
        self.field_kwargs = kwargs

        super().__init__(**{})

    def sub_field_contribute_to_class(
        self,

        cls,
        field_name,

        sub_field_name,
        sub_filed_class,
        sub_field_default_attr=None,
    ):

        my_kwargs = self.__get_my_kwargs(
            my_name=sub_field_name,
            kwargs=self.field_kwargs,
            default=sub_field_default_attr,
        )

        sub_field = sub_filed_class(
            name=field_name + f'_{sub_field_name}',
            **my_kwargs,
        )

        sub_field.contribute_to_class(cls, sub_field.name)

        setattr(sub_field, 'PARENT_FIELD_NAME', field_name)

    @staticmethod
    def __get_my_kwargs(my_name, kwargs: dict, default: dict = None):
        kwargs_ = {}
        for k, v in kwargs.items():
            if f'{my_name}__' in k:
                kwargs_[k[len(f'{my_name}__'):]] = v

        default_ = default or {}
        for k, v in default_.items():
            if k not in kwargs_:
                kwargs_[k] = v

        return kwargs_


class Image:
    def __init__(self, **kwargs):
        self.category = kwargs.get('category')
        self.device_type = kwargs.get('device_type')
        self.alt = kwargs.get('alt')
        self.description = kwargs.get('description')
        self.width = kwargs.get('width')
        self.height = kwargs.get('height')
        self.src = kwargs.get('src')
        self.blurbase64 = kwargs.get('blurbase64')


class ImageProperty:
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        if not instance:
            return self

        ret = Image(
            category=getattr(instance, self.name + '_category'),
            device_type=getattr(instance, self.name + '_device_type'),
            alt=getattr(instance, self.name + '_alt'),
            description=getattr(instance, self.name + '_description'),
            width=getattr(instance, self.name + '_width'),
            height=getattr(instance, self.name + '_height'),
            src=getattr(instance, self.name + '_src'),
            blurbase64=getattr(instance, self.name + '_blurbase64'),
        )
        return ret

    def __set__(self, instance, value: dict):
        src = value.get('src')

        blurbase64 = (small_blur_base64(src) if src.name.split('.')[-1] != 'svg' else None) if src else None

        setattr(
            instance, self.name + '_category', value.get('category'))
        setattr(
            instance, self.name + '_device_type', value.get('device_type'))
        setattr(
            instance, self.name + '_alt', value.get('alt'))
        setattr(
            instance, self.name + '_description', value.get('description'))
        setattr(
            instance, self.name + '_width', value.get('width'))
        setattr(
            instance, self.name + '_height', value.get('height'))
        setattr(
            instance, self.name + '_src', src)
        setattr(
            instance, self.name + '_blurbase64', blurbase64)


class ImageField(BaseMultiplyField):

    def contribute_to_class(self, cls, name, private_only=False):
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='category',
            sub_filed_class=models.IntegerField,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} Category Name'),
                'help_text': _('In this field, we specify in which category this image is classified.'),
                'blank': False,
                'null': False,
                'choices': (
                    (1, 'Unknown',),
                    (2, 'Image',),
                    (3, 'Avatar',),
                    (4, 'Icon',),
                ),
                'default': 1,
            }
        )
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='device_type',
            sub_filed_class=models.IntegerField,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} Device Type'),
                'help_text': _('In this field, we specify which device this image is suitable for.'),
                'blank': False,
                'null': False,
                'choices': (
                    (1, 'Mobile',),
                    (2, 'Desktop',),
                    (3, 'Both',),
                ),
                'default': 3,
            }
        )
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='alt',
            sub_filed_class=models.CharField,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} Alternate Title'),
                'help_text': _('This title will be shown instead if the image is not loaded.'),
                'max_length': 255,
                'blank': True,
                'null': True,
                'default': 'beensi',
            }
        )
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='description',
            sub_filed_class=models.TextField,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} Description'),
                'help_text': _("If you want a text to be written below the image,"
                               " write it here. Otherwise, leave it blank."),
                'blank': True,
                'null': True,
                'default': 'beensi',
            }
        )
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='width',
            sub_filed_class=models.PositiveIntegerField,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} Width'),
                'blank': True,
                'null': True,
                'default': 2400,
            }
        )
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='height',
            sub_filed_class=models.PositiveIntegerField,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} Height'),
                'blank': True,
                'null': True,
                'default': 1600,
            }
        )
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='src',
            sub_filed_class=models.ImageField,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} SRC'),
                'upload_to': ImageUploadTo(),
                'blank': True,
                'null': True,

                'width_field': f'{name}_width',
                'height_field': f'{name}_height',
            }
        )
        self.sub_field_contribute_to_class(
            cls,
            field_name=name,
            sub_field_name='blurbase64',
            sub_filed_class=BlurBase64Field,
            sub_field_default_attr={
                'verbose_name': _(f'{verbose_name(name)} Blur-Base64'),
                'blank': True,
                'null': True,
            }
        )

        setattr(cls, name, ImageProperty(name))


@deconstructible
class PhoneNumberValidator(validators.RegexValidator):
    regex = r"^([+])([0-9]){1,31}$"
    message = _(
        "Enter a valid phone_number.\n"
        "Sample +123456789012"
    )
    flags = 0


class PhoneNumberField(models.CharField):
    __validator = PhoneNumberValidator()
    default_validators = [__validator]
    description = _('PhoneNumber')

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 31)
        super().__init__(*args, **kwargs)


@deconstructible
class UnicodeUsernameValidator(validators.RegexValidator):
    regex = r"^([0-9a-zA-Z_+]){3,127}$"
    message = _(
        "Enter a valid username. This value may contain only letters, "
        "numbers, and _,+ character."
    )
    flags = 0


class UsernameField(models.CharField):
    __validator = UnicodeUsernameValidator()
    default_validators = [__validator]
    description = _('Username')

    def __init__(self, *args, **kwargs):
        kwargs.setdefault(
            'max_length', 127, )
        kwargs.setdefault(
            'help_text', _('Required. 255 characters or fewer. Letters, digits and @/./+/-/_ only.'), )
        kwargs.setdefault(
            'error_messages', {'unique': _('A user with that username already exists.')}, )

        super().__init__(*args, **kwargs)


@deconstructible
class PasswordValidator(validators.RegexValidator):
    regex = r"^(?=.*\d)(?=.*[a-z])(?=.*[A-Z])(?=.*[!@#$%^&*_]).{8,31}$"
    message = _(
        "Enter a valid password. This value may contain only letters, "
        "numbers, and characters(! @ # $ % ^ & * _)."
    )
    flags = 0


class PasswordField(models.CharField):
    __validator = PasswordValidator()
    default_validators = [__validator]
    description = _('Password')

    def __init__(self, *args, **kwargs):
        kwargs.setdefault(
            'max_length', 255, )

        super().__init__(*args, **kwargs)


class BooleanNumberField(models.IntegerField):
    description = _('Boolean Number')

    def __init__(self, *args, **kwargs):
        kwargs.setdefault(
            'default', 0, )
        kwargs.setdefault(
            'choices', ((0, 'No',), (1, 'Yes',),), )

        super().__init__(*args, **kwargs)



from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError


def validate(value):
    validation_error_message = _('Please enter a valid number.')

    value = str(value).lower()
    match value:
        case 't':
            return True
        case 'true':
            return True
        case '1':
            return True
        case 'f':
            return False
        case 'false':
            return False
        case '0':
            return False
        case _:
            raise ValidationError(validation_error_message)



from gando.architectures.services import BaseDataBaseManagerService


class BaseCreatorService(BaseDataBaseManagerService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.valid_input_data = self.__validate(**kwargs)

    @property
    def valid_key_input_data(self) -> list:
        if not hasattr(self, 'valid_key_input_data_list'):
            raise Exception('valid_key_input_data_list not define.')
        return self.valid_key_input_data_list

    def __validate(self, **kwargs):
        tmp = {}
        for k, v in kwargs.items():
            if k in self.valid_key_input_data:
                tmp[k] = v
        ret = tmp
        return ret

    def __create_db_record(self):
        try:
            obj = self.model.objects.create(**self.valid_input_data)
        except Exception as exc:
            raise exc
        return obj.__dict__

    def service_output_handler(self, *args, **kwargs):
        return self.__create_db_record()



import re

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

from datetime import datetime, date


def validate(value):
    pattern = r'^(?=.*\d)(?=.*[a-z])(?=.*[A-Z])(?=.*[a-zA-Z]).{8,31}$'
    validation_error_message = _(
        "Please enter your password as a combination of "
        "uppercase English letters and "
        "lowercase English letters, "
        "English numbers and characters (!#*_-.).\n"
        "The minimum number of characters is 8 and the maximum is 31.")

    value = str(value)
    if not re.match(pattern, value):
        raise ValidationError(validation_error_message)
    return value



import re

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError


def validate(value):
    pattern = r'^([0]){1}$|^([1-9]){1}([0-9])*$|^([0]){1}([.])([0-9])+$|^([1-9]){1}([0-9])*([.])([0-9])+$'
    validation_error_message = _('Please enter an integer or a decimal number.')

    value = str(value)
    if not re.match(pattern, value):
        raise ValidationError(validation_error_message)

    value = float(value)
    return value



from gando.config import SETTINGS
from django.utils.deprecation import MiddlewareMixin


def _get_user_agent_device_key(key):
    from gando.models import UserAgentDevice

    if not key:
        return None
    try:
        obj = UserAgentDevice.objects.get(key=key)
        return obj.key
    except:
        return None


def _create_user_agent_device_key(request):
    from gando.models import UserAgentDevice, user_agent_device_key_creator
    from gando.utils.client import UADSchema

    params = UADSchema(request)
    try:
        obj = UserAgentDevice.objects.create(
            **params.model_dump()
        )
        return obj.key
    except:
        key = user_agent_device_key_creator(params)
        return key


def user_agent_device_id(request):
    if SETTINGS.USER_AGENT_DEVICE_HANDLER.HANDLING is False:
        return None

    cookie_name = SETTINGS.USER_AGENT_DEVICE_HANDLER.COOKIE_NAME
    if not hasattr(request, 'COOKIES'):
        return None

    adk = request.COOKIES.get(cookie_name)

    uad_key = _get_user_agent_device_key(adk)
    if uad_key:
        return uad_key

    uad_key = _create_user_agent_device_key(request)
    return uad_key


class UserAgentDeviceMiddleware(MiddlewareMixin):

    def process_request(self, request):
        setattr(request, 'uad', user_agent_device_id(request))

    def process_response(self, request, response):
        self.process_request(request)

        if SETTINGS.USER_AGENT_DEVICE_HANDLER.HANDLING is False:
            return response

        response.set_cookie(
            key=SETTINGS.USER_AGENT_DEVICE_HANDLER.COOKIE_NAME,
            value=request.uad,
            max_age=60 * 60 * 24 * 365,
            httponly=False,
            secure=False
        )
        return response



from typing import OrderedDict
from rest_framework.serializers import ModelSerializer as DRFModelSerializer


class ModelSerializer(DRFModelSerializer):
    def to_dict(self):
        data = self.data
        data_ = self._to_dict(data)
        return data_

    def _to_dict(self, data):
        if isinstance(data, list):
            data_ = []
            for i in data:
                data_.append(self._to_dict(i))
            return data_

        if isinstance(data, dict) or isinstance(data, OrderedDict):
            data = dict(data)
            data_ = {}
            for k, v in data.items():
                data_[k] = self._to_dict(v)
            return data_

        if isinstance(data, str) or isinstance(data, int) or isinstance(data, float) or isinstance(data, bool):
            data_ = data
            return data_

        data_ = str(data)
        return data_



from gando.architectures.services import BaseService
from django.db.models import Model
from abc import abstractmethod
from pydantic import BaseModel


class BaseDataBaseManagerService(BaseService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def model(self) -> Model:
        if not hasattr(self, 'model_class'):
            raise Exception('model_class not define.')
        return self.model_class

    @property
    def output_schema(self) -> BaseModel:
        if not hasattr(self, 'output_schema_class'):
            raise Exception('output_schema_class not define.')
        return self.output_schema_class

    def __convert_to_schema(self, obj):
        ret = self.output_schema(**obj) if obj else None
        return ret

    def execute(self, *args, **kwargs):
        rslt = super().execute(*args, **kwargs)
        ret = self.__convert_to_schema(rslt)
        return ret

    @abstractmethod
    def service_output_handler(self, *args, **kwargs):
        pass



from pydantic import BaseModel
from typing import List

from uuid import UUID


class ObjectSchema(BaseModel):
    id: UUID
    service_name: str
    app_name: str
    table_name: str


class RelationsSchema(BaseModel):
    objs: List[ObjectSchema]



class BaseMessage:
    def __init__(self, message, code, type_, *args, **kwargs):
        self.message = message
        self.code = code
        self.type = type_


class BaseFailMessage(BaseMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(type_='FAIL', *args, **kwargs)


class BaseWarningMessage(BaseMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(type_='WARNING', *args, **kwargs)


class BaseSuccessMessage(BaseMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(type_='SUCCESS', *args, **kwargs)


class BaseInfoMessage(BaseMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(type_='INFO', *args, **kwargs)


class _DefaultResponse100FailMessage(BaseFailMessage):
    def __init__(self):
        super().__init__(
            message='please wait...',
            code='100',
        )


DefaultResponse100FailMessage = _DefaultResponse100FailMessage()


class _DefaultResponse200SuccessMessage(BaseSuccessMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='Your request has been successfully registered.',
            code='200',
        )


DefaultResponse200SuccessMessage = _DefaultResponse200SuccessMessage()


class _DefaultResponse201SuccessMessage(BaseSuccessMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='The desired object was created correctly.',
            code='201',
        )


DefaultResponse201SuccessMessage = _DefaultResponse201SuccessMessage()


class _DefaultResponse300FailMessage(BaseFailMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='The requirements for your request are not available.',
            code='300',
        )


DefaultResponse300FailMessage = _DefaultResponse300FailMessage()


class _DefaultResponse400FailMessage(BaseFailMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='Bad Request...',
            code='400',
        )


DefaultResponse400FailMessage = _DefaultResponse400FailMessage()


class _DefaultResponse401FailMessage(BaseFailMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='Your authentication information is not available.',
            code='401',
        )


DefaultResponse401FailMessage = _DefaultResponse401FailMessage()


class _DefaultResponse403FailMessage(BaseFailMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='You do not have access to this section.',
            code='403',
        )


DefaultResponse403FailMessage = _DefaultResponse403FailMessage()


class _DefaultResponse404FailMessage(BaseFailMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='There is no information about your request.',
            code='404',
        )


DefaultResponse404FailMessage = _DefaultResponse404FailMessage()


class _DefaultResponse421FailMessage(BaseFailMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message=(
                "An unexpected error has occurred based on your request type.\n"
                "Please do not repeat this request without changing your request.\n"
                "Be sure to read the documents on how to use this service correctly.\n"
                "In any case, discuss the issue with support.\n"
            ),
            code='421',
        )


DefaultResponse421FailMessage = _DefaultResponse421FailMessage()


class _DefaultResponse500FailMessage(BaseFailMessage):
    def __init__(self, *args, **kwargs):
        super().__init__(
            message='The server is unable to respond to your request.',
            code='500',
        )


DefaultResponse500FailMessage = _DefaultResponse500FailMessage()



from pydantic import BaseModel
from typing import List


class Widget(BaseModel):
    name: str
    priority: int
    attributes: dict


class Section(BaseModel):
    name: str
    priority: int
    widgets: List[Widget]


class Structure(BaseModel):
    name: str
    sections: List[Section]
    meta: dict = {}



import re

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError
from django.utils.timezone import now as dj_now

from datetime import datetime, time


def validate(value):
    pattern = r'^([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9])$'
    validation_error_message = _('Entered in the following format.\nhh:mm:ss')

    value = str(value)
    if not re.match(pattern, value):
        raise ValidationError(validation_error_message)

    n = dj_now()
    n = f'{n.year}-{n.month}-{n.day}'
    value = datetime.strptime(f'{n} {value}', '%Y-%m-%d %H:%M:%S')
    value = time(hour=value.hour, minute=value.minute, second=value.second)
    return value
